"""Parsing the Key-Value-Pair (KVP) request format."""

from __future__ import annotations

from copy import copy

from gisserver.exceptions import (
    InvalidParameterValue,
    MissingParameterValue,
    OperationParsingFailed,
    wrap_parser_errors,
)
from gisserver.parsers.xml import parse_qname, xmlns

REQUIRED = ...  # sentinel value


class KVPRequest:
    """The Key-Value-Pair (KVP) request format.

    This handles parameters from the HTTP GET request.
    It includes notation format support for certain parameters,
    such as comma-separated lists, and parenthesis-grouping notations.

    Some basic validation is performed, allowing to convert the data into Python types.
    """

    def __init__(self, query_string: dict[str, str], ns_aliases: dict[str, str] | None = None):
        # The parameters are case-insensitive.
        self.params = {name.upper(): value for name, value in query_string.items()}

        # Make sure most common namespaces are known for resolving them.
        self.ns_aliases = {
            **xmlns.as_ns_aliases(),  # some defaults in case these are missing in the request
            **(ns_aliases or {}),  # our local application namespaces
            **parse_kvp_namespaces(self.params.get("NAMESPACES")),  # extra in the request
        }

    def __contains__(self, name: str) -> bool:
        """Tell whether a parameter is present."""
        return name.upper() in self.params

    def get_custom(self, name: str, *, alias: str | None = None, default=REQUIRED, parser=None):
        """Retrieve a value by name.
        This performs basic validation, similar to what the XML parsing does.

        Any parsing errors or validation checks are raised as WFS Exceptions,
        meaning the client will get the appropriate response.

        :param name: The name of the parameter, typically given in its XML notation format (camelCase).
        :param alias: An older WFS 1 to try for compatibility (e.g. TYPENAMES/TYPENAME, COUNT/MAXFEATURES)
        :param default: The default value to return. If not provided, the parameter is required.
        :param parser: A custom Python function or type to convert the value with.
        """
        kvp_name = name.upper()
        value = self.params.get(kvp_name)
        if not value and alias:
            value = self.params.get(alias.upper())

        # Check required field settings, both empty and missing value are treated the same.
        if not value:
            if default is REQUIRED:
                if value is None:
                    raise MissingParameterValue(
                        f"Missing required '{name}' parameter.", locator=name
                    )
                else:
                    raise InvalidParameterValue(f"Empty '{kvp_name}' parameter", locator=name)
            return default

        # Allow conversion into a python object
        if parser is not None:
            with wrap_parser_errors(kvp_name, locator=name):
                return parser(value)
        else:
            return value

    def get_str(
        self, name: str, *, alias: str | None = None, default: str | None = REQUIRED
    ) -> str | None:
        """Retrieve a string value from the request."""
        return self.get_custom(name, alias=alias, default=default)

    def get_int(
        self, name: str, *, alias: str | None = None, default: int | None = REQUIRED
    ) -> int | None:
        """Retrieve an integer value from the request."""
        return self.get_custom(name, alias=alias, default=default, parser=int)

    def get_list(
        self, name, *, alias: str | None = None, default: list | None = REQUIRED
    ) -> list[str] | None:
        """Retrieve a comma-separated list value from the request."""
        return self.get_custom(
            name,
            alias=alias,
            default=default,
            parser=lambda x: x.split(","),
        )

    def parse_qname(self, value) -> str:
        """Convert the value to an XML fully qualified name."""
        return parse_qname(value, self.ns_aliases)

    def split_parameter_lists(self) -> list[KVPRequest]:
        """Split the parameter lists into individual requests.

        This translates a request such as:

        .. code-block:: urlencoded

            TYPENAMES=(ns1:F1,ns2:F2)(ns1:F1,ns1:F1)
            &ALIASES=(A,B)(C,D)
            &FILTER=(<Filter>… for A,B …</Filter>)(<Filter>…for C,D…</Filter>)

        into separate pairs:

        .. code-block:: urlencoded

            TYPENAMES=ns1:F1,ns2:F2&ALIASES=A,B&FILTER=<Filter>…for A,B…</Filter>
            TYPENAMES=ns1:F1,ns1:F1&ALIASES=C,D&FILTER=<Filter>…for C,D…</Filter>

        It's both possible have some query parameters split and some shared.
        For example to have two different bounding boxes:

        .. code-block:: urlencoded

             TYPENAMES=(INWATER_1M)(BuiltUpA_1M)&BBOX=(40.9821,...)(40.5874,...)

        or have a single bounding box for both queries:

        .. code-block:: urlencoded

             TYPENAMES=(INWATER_1M)(BuiltUpA_1M)&BBOX=40.9821,23.4948,41.0257,23.5525
        """
        pairs = {
            name: value[1:-1].split(")(")
            for name, value in self.params.items()
            if value.startswith("(") and value.endswith(")")
        }
        if not pairs:
            return [self]

        pair_sizes = {len(value) for value in pairs.values()}
        if len(pair_sizes) > 1:
            keys = sorted(pairs)
            raise OperationParsingFailed(
                f"Inconsistent pairs between: {', '.join(keys)}", locator=keys[0]
            )

        # Produce variations of the same request object
        pair_size = next(iter(pair_sizes))
        variants = []
        for i in range(pair_size):
            updates = {key: value[i] for key, value in pairs.items()}
            variant = copy(self)
            variant.params = {**self.params, **updates}
            variants.append(variant)
        return variants


def parse_kvp_namespaces(value) -> dict[str, str]:
    """Parse the 'NAMESPACES' parameter format to lookups.

    The NAMESPACES parameter defines which namespaces are used in the KVP request.
    When this parameter is not given, the default namespaces are assumed.
    """
    if not value:
        return {}

    # example single value: xmlns(http://example.org)
    # or: NAMESPACES=xmlns(xml,http://www.w3.org/...),xmlns(wfs,http://www.opengis.net/...)
    tokens = value.split(",")

    namespaces = {}
    tokens = iter(tokens)
    for prefix in tokens:
        if not prefix.startswith("xmlns("):
            raise InvalidParameterValue(
                f"Expected xmlns(...) format: {value}", locator="namespaces"
            )
        if prefix.endswith(")"):
            # xmlns(http://...)
            uri = prefix[6:-1]
            prefix = ""
        else:
            uri = next(tokens, "")
            if not uri.endswith(")"):
                raise InvalidParameterValue(
                    f"Expected xmlns(prefix,uri) format: {value}",
                    locator="namespaces",
                )
            prefix = prefix[6:]
            uri = uri[:-1]

        namespaces[prefix] = uri

    return namespaces



"""Wrappers for the results of GetFeature/GetPropertyValue.

The "SimpleFeatureCollection" and "FeatureCollection" and their
properties match the WFS 2.0 spec closely.
"""

from __future__ import annotations

import math
import typing
from collections.abc import Iterable
from datetime import timezone
from functools import cached_property

from django.db import models
from django.utils.timezone import now

from gisserver import conf
from gisserver.exceptions import wrap_filter_errors
from gisserver.features import FeatureType

from .iters import ChunkedQuerySetIterator, CountingIterator

if typing.TYPE_CHECKING:
    from gisserver.projection import FeatureProjection, QueryExpression


CALCULATE = -9999999


class SimpleFeatureCollection:
    """Wrapper to read a result set.

    This object type is defined in the WFS spec.
    It holds a collection of ``<wfs:member>`` objects.
    """

    def __init__(
        self,
        source_query: QueryExpression,
        feature_types: list[FeatureType],
        queryset: models.QuerySet,
        start: int,
        stop: int,
        number_matched: int | None = CALCULATE,
    ):
        self.source_query = source_query
        self.feature_types = feature_types
        self.queryset = queryset
        self.start = start
        self.stop = stop
        self._number_matched = number_matched

        self._result_cache = None
        self._result_iterator = None
        self._has_more = None

        # Tell that is a resultType=hits request.
        # Typically, start and stop are 0. However, for resultType=hits with count,
        # that does not apply. Instead, it detects whether the known amount is already provided.
        # Detecting that queryset.none() is provided won't work, as that can be used by IdOperator too.
        self._is_hits_request = number_matched is not None and number_matched != CALCULATE

    def __iter__(self) -> Iterable[models.Model]:
        """Iterate through all results.

        Depending on the output format (and whether pagination is read first),
        the results can either be cached first, or be streamed without caching.
        This picks the best-performance scenario in most cases.
        """
        if self._is_hits_request:
            self._result_cache = []

        if self._result_cache is not None:
            # Repeat what's cached if reading again.
            return iter(self._result_cache)
        elif self.queryset._prefetch_related_lookups:
            # Optimal solution to have an iterator that also handles prefetches
            return self._chunked_iterator()
        else:
            # No need to fetch everything first, can just stream the results.
            return self.iterator()

    def iterator(self):
        """Explicitly request the results to be streamed.

        This can be used by output formats that stream results, and don't
        access :attr:`number_returned`.
        Note this is not compatible with ``prefetch_related()``.
        """
        if self._result_iterator is not None:
            raise RuntimeError("Results for feature collection are read twice.")

        if self._result_cache is not None:
            # In case the results were read already, reuse that.
            return iter(self._result_cache)
        elif self._is_hits_request:
            # resulttype=hits
            return iter([])
        else:
            if self._use_sentinel_record:
                model_iter = self._paginated_queryset(add_sentinel=True).iterator()
                self._result_iterator = CountingIterator(
                    model_iter, max_results=(self.stop - self.start)
                )
            else:
                model_iter = self._paginated_queryset().iterator()
                self._result_iterator = CountingIterator(model_iter)
            return iter(self._result_iterator)

    def _chunked_iterator(self):
        """Generate an interator that processes results in chunks."""
        # Private function so the same logic of .iterator() is not repeated.
        self._result_iterator = ChunkedQuerySetIterator(self._paginated_queryset())
        return iter(self._result_iterator)

    def _paginated_queryset(self, add_sentinel=True) -> models.QuerySet:
        """Apply the pagination to the queryset."""
        if self.stop == math.inf:
            # Infinite page requested
            if self.start:
                return self.queryset[self.start :]
            else:
                return self.queryset
        else:
            return self.queryset[self.start : self.stop + (1 if add_sentinel else 0)]

    def first(self):
        with wrap_filter_errors(self.source_query):
            try:
                # Don't query a full page, return only one instance (for GetFeatureById)
                # This also preserves the extra added annotations (like _as_gml_FIELD)
                return self.queryset[self.start]
            except IndexError:
                return None

    def fetch_results(self):
        """Forcefully read the results early."""
        if self._result_cache is not None:
            return
        if self._result_iterator is not None:
            raise RuntimeError("Results for feature collection are read twice.")

        if self._is_hits_request:
            self._result_cache = []
        else:
            # This still allows prefetch_related() to work,
            # since QuerySet.iterator() is avoided.
            if self.stop == math.inf:
                # Infinite page requested, see if start is still requested
                qs = self.queryset[self.start :] if self.start else self.queryset.all()

                with wrap_filter_errors(self.source_query):
                    self._result_cache = list(qs)
            elif self._use_sentinel_record:
                # No counting, but instead fetch an extra item as sentinel to see if there are more results.
                qs = self.queryset[self.start : self.stop + 1]

                with wrap_filter_errors(self.source_query):
                    page_results = list(qs)

                # The stop + 1 sentinel allows checking if there is a next page.
                # This means no COUNT() is needed to detect that.
                page_size = self.stop - self.start
                self._has_more = len(page_results) > page_size
                if self._has_more:
                    # remove extra element
                    page_results.pop()

                self._result_cache = page_results
            else:
                # Fetch exactly the page size, no more is needed.
                # Will use a COUNT on the total table, so it can be used to see if there are more pages.
                qs = self.queryset[self.start : self.stop]

                with wrap_filter_errors(self.source_query):
                    self._result_cache = list(qs)

    @cached_property
    def _use_sentinel_record(self) -> bool:
        """Tell whether a sentinel record should be included in the result set.
        This is used to determine whether there are more results, without having to perform a COUNT query
        """
        return conf.GISSERVER_COUNT_NUMBER_MATCHED == 0 or (
            conf.GISSERVER_COUNT_NUMBER_MATCHED == 2 and self.start
        )

    @cached_property
    def number_returned(self) -> int:
        """Return the number of results for this page."""
        if self._is_hits_request:
            return 0
        elif self._result_iterator is not None:
            # When requesting the data after the fact, results are counted.
            return self._result_iterator.number_returned
        else:
            # Count by fetching all data. Otherwise, the results are queried twice.
            # For GML/XML, it's not possible the stream the queryset results
            # as the first tag needs to describe the number of results.
            self.fetch_results()
            return len(self._result_cache)

    @property
    def number_matched(self) -> int:
        """Return the total number of matches across all pages."""
        if self._is_hits_request:
            if self.stop:
                # resulttype=hits&COUNT=n should minimize how many are "matched".
                return min(self._number_matched, self.stop - self.start)
            else:
                return self._number_matched
        elif self._number_matched != CALCULATE:
            # Return previously cached result
            return self._number_matched

        if self._is_surely_last_page:
            # For resulttype=results, an expensive COUNT query can be avoided
            # when this is the first and only page or the last page.
            return self.start + self.number_returned

        qs = self.queryset
        clean_annotations = {
            # HACK: remove database optimizations from output renderer.
            # Otherwise, it becomes SELECT COUNT(*) FROM (SELECT AsGML(..), ...)
            key: value
            for key, value in qs.query.annotations.items()
            if not key.startswith("_as_") and not key.startswith("_As")  # AsGML / AsEWKT
        }
        if clean_annotations != qs.query.annotations:
            qs = self.queryset.all()  # make a clone to allow editing
            qs.query.annotations = clean_annotations

        # Calculate, cache and return
        with wrap_filter_errors(self.source_query):
            self._number_matched = qs.count()
        return self._number_matched

    @property
    def _is_surely_last_page(self):
        """Return true when it's totally clear this is the last page."""
        if self.start == self.stop == 0:
            return True  # hits request without count
        elif self._is_hits_request:
            return False

        # Optimization to avoid making COUNT() queries when we can already know the answer.
        if self.stop == math.inf:
            return True  # Infinite page requested
        elif self._use_sentinel_record:
            if self._has_more is not None:
                # did page+1 record check here, answer is known.
                return not self._has_more
            elif (
                isinstance(self._result_iterator, CountingIterator)
                and self._result_iterator.has_more is not None
            ):
                # did page+1 record check via the CountingIterator.
                return not self._result_iterator.has_more

        # Here different things will happen.
        # For GeoJSON output, the iterator was read first, and `number_returned` is already filled in.
        # For GML output, the pagination details are requested first, and will fetch all data.
        # Hence, reading `number_returned` here can be quite an intensive operation.
        page_size = self.stop - self.start
        return page_size and (self.number_returned < page_size or self._has_more is False)

    @property
    def has_next(self):
        if self.stop == math.inf or (self.start == self.stop == 0):
            return False
        elif self._has_more is not None:
            return self._has_more  # did page+1 record check, answer is known.
        elif self._is_surely_last_page:
            return False  # Fewer results than expected, answer is known.

        if self._is_hits_request:
            return self.stop <= self._number_matched
        else:
            # This will perform an slow COUNT() query...
            return self.stop < self.number_matched

    @cached_property
    def projection(self) -> FeatureProjection:
        """Provide the projection to render these results with."""
        # Note this attribute would technically be part of the 'query' object,
        # but since the projection needs to be calculated once, it's stored here for convenience.
        return self.source_query.get_projection()


class FeatureCollection:
    """WFS object that holds the result type for ``GetFeature``.
    This object type is defined in the WFS spec.
    It holds a collection of :class:`SimpleFeatureCollection` results.
    """

    def __init__(
        self,
        results: list[SimpleFeatureCollection],
        number_matched: int | None = CALCULATE,
        next: str | None = None,
        previous: str | None = None,
    ):
        """
        :param source_query: The query that generated this output.
        :param results: All retrieved feature collections (one per FeatureType)
        :param number_matched: Total number of features across all pages
        :param next: URL of the next page
        :param previous: URL of the previous page
        """
        self.results = results
        self._number_matched = number_matched
        self.next = next
        self.previous = previous
        self.date = now()
        self.timestamp = self.date.astimezone(timezone.utc).isoformat()

    @cached_property
    def number_returned(self) -> int:
        """Return the total number of returned features"""
        return sum(c.number_returned for c in self.results)

    @cached_property
    def number_matched(self) -> int | None:
        """The number of features matched, None means "unknown"."""
        if self._number_matched is None:
            return None  # WFS allows returning this as "unknown"
        elif self._number_matched == CALCULATE:
            # By making the number_matched lazy, the calling method has a chance to
            # decorate the results with extra annotations before they are evaluated.
            # This is needed since SimpleCollection.number_matched already evaluates the queryset.
            if conf.GISSERVER_COUNT_NUMBER_MATCHED == 0 or (
                conf.GISSERVER_COUNT_NUMBER_MATCHED == 2 and self.results[0].start > 0
            ):
                # Report "unknown" for either all pages, or the second page.
                # Most clients don't need this metadata, and thus we avoid a COUNT query.
                return None

            return sum(c.number_matched for c in self.results)
        else:
            # Evaluate any lazy attributes
            return int(self._number_matched)

    @property
    def has_next(self) -> bool:
        """Efficient way to see if a next link needs to be written.

        This method will show up in profiling through
        as it will be the first moment where queries are executed.
        """
        if all(c._is_surely_last_page for c in self.results):
            return False

        for c in self.results:  # noqa: SIM110
            # This may perform an COUNT query or read the results and detect the sentinel object:
            if c.has_next:
                return True
        return False

    def __iter__(self):
        return iter(self.results)



"""Outputting XML for the stored query logic."""

from __future__ import annotations

from io import StringIO
from xml.etree.ElementTree import Element, tostring

from gisserver.extensions.queries import QueryExpressionText, StoredQueryDescription
from gisserver.output.utils import attr_escape, tag_escape
from gisserver.parsers.xml import xmlns

from .base import XmlOutputRenderer


class ListStoredQueriesRenderer(XmlOutputRenderer):
    """Rendering for the ``<wfs:ListStoredQueriesResponse>``."""

    # XML Namespaces to include by default
    xml_namespaces = {
        xmlns.wfs20: "",
        xmlns.xs: "xs",
        xmlns.xsi: "xsi",
    }

    def __init__(self, operation, query_descriptions: list[StoredQueryDescription]):
        """Take the list of stored queries to render."""
        super().__init__(operation)
        self.all_feature_types = operation.view.get_bound_feature_types()
        self.query_descriptions = query_descriptions

    def render_stream(self):
        self.output = StringIO()
        self.output.write(
            '<?xml version="1.0" encoding="UTF-8"?>\n'
            f"<ListStoredQueriesResponse"
            f" {self.render_xmlns_attributes()}"
            f' xsi:schemaLocation="http://www.opengis.net/wfs/2.0 http://schemas.opengis.net/wfs/2.0/wfs.xsd">\n'
        )
        for query_description in self.query_descriptions:
            self.write_query(query_description)

        self.output.write("</ListStoredQueriesResponse>\n")
        return self.output.getvalue()

    def write_query(self, query_description: StoredQueryDescription):
        self.output.write(
            f'  <StoredQuery id="{query_description.id}">\n'
            f"    <Title>{tag_escape(query_description.title)}</Title>\n"
        )

        for expression in query_description.expressions:
            return_types = expression.return_feature_types or self.all_feature_types
            for return_type in return_types:
                feature_qname = self.feature_to_qname(return_type)
                self.output.write(
                    f"    <ReturnFeatureType>{tag_escape(feature_qname)}</ReturnFeatureType>\n"
                )

        self.output.write("  </StoredQuery>\n")


class DescribeStoredQueriesRenderer(XmlOutputRenderer):
    """Rendering for the ``<wfs:DescribeStoredQueriesResponse>``."""

    # XML Namespaces to include by default
    xml_namespaces = {
        xmlns.wfs20: "",
        xmlns.xs: "xs",
        xmlns.xsi: "xsi",
    }

    def __init__(self, operation, query_descriptions: list[StoredQueryDescription]):
        """Take the list of stored queries to render."""
        super().__init__(operation)
        self.all_feature_types = operation.view.get_bound_feature_types()
        self.query_descriptions = query_descriptions

    def render_stream(self):
        self.output = StringIO()
        self.output.write(
            '<?xml version="1.0" encoding="UTF-8"?>\n'
            f"<DescribeStoredQueriesResponse"
            f" {self.render_xmlns_attributes()}"
            f' xsi:schemaLocation="http://www.opengis.net/wfs/2.0 http://schemas.opengis.net/wfs/2.0/wfs.xsd">\n'
        )

        for query_description in self.query_descriptions:
            self.write_description(query_description)

        self.output.write("</DescribeStoredQueriesResponse>\n")
        return self.output.getvalue()

    def write_description(self, query_description: StoredQueryDescription):
        """Write the stored query description."""
        self.output.write(
            f'<StoredQueryDescription id="{attr_escape(query_description.id)}">\n'
            f"  <Title>{tag_escape(query_description.title)}</Title>\n"
            f"  <Abstract>{tag_escape(query_description.abstract)}</Abstract>\n"
        )

        # Declare parameters
        for name, xsd_type in query_description.parameters.items():
            type_qname = self.to_qname(xsd_type)
            self.output.write(f'  <Parameter name="{attr_escape(name)}" type="{type_qname}"/>\n')

        # The QueryExpressionText nodes allow code per return type.
        for expression in query_description.expressions:
            self.render_expression(expression)

        self.output.write("</StoredQueryDescription>\n")

    def render_expression(self, expression: QueryExpressionText):
        """Render the 'QueryExpressionText' node (no body content for now)."""
        is_private = "true" if expression.is_private else "false"
        if expression.return_feature_types is None:
            # for GetFeatureById
            types = " ".join(self.feature_to_qname(ft) for ft in self.all_feature_types)
        else:
            types = " ".join(
                self.feature_to_qname(return_type)
                for return_type in expression.return_feature_types
            )

        if expression.is_private or not expression.implementation_text:
            implementation_text = ""
        elif isinstance(expression.implementation_text, Element):
            # XML serialization (will recreate namespaces)
            default_namespace = next(
                (ns for ns, prefix in self.app_namespaces.items() if prefix == ""), None
            )
            implementation_text = tostring(
                expression.implementation_text,
                xml_declaration=False,
                default_namespace=default_namespace,
            )
        else:
            # Some raw content (e.g. language="python")
            implementation_text = tag_escape(expression.implementation_text)

        self.output.write(
            f'  <QueryExpressionText isPrivate="{is_private}" language="{expression.language}"'
            f' returnFeatureTypes="{types}">{implementation_text}</QueryExpressionText>\n'
        )



"""The view layer parses the request, and dispatches it to an operation."""

from __future__ import annotations

import logging
import re
from urllib.parse import unquote_plus, urlencode

from django.core.exceptions import ImproperlyConfigured, SuspiciousOperation
from django.core.exceptions import PermissionDenied as Django_PermissionDenied
from django.shortcuts import render
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from gisserver import conf
from gisserver.exceptions import (
    ExternalParsingError,
    InvalidParameterValue,
    OperationNotSupported,
    OperationParsingFailed,
    OperationProcessingFailed,
    OWSException,
    PermissionDenied,
    XmlElementNotSupported,
)
from gisserver.features import FeatureType, ServiceDescription
from gisserver.operations import base, wfs20
from gisserver.parsers.ows import (
    BaseOwsRequest,
    KVPRequest,
    resolve_kvp_parser_class,
    resolve_xml_parser_class,
)
from gisserver.parsers.xml import parse_xml_from_string, split_ns

logger = logging.getLogger(__name__)

SAFE_VERSION = re.compile(r"\A[0-9.]+\Z")


class OWSView(View):
    """The base logic to implement OGC view like WFS.

    Each subclass defines 'accept_operations' with the desired RPC operations.
    """

    #: Define the namespace to use in the XML
    xml_namespace = "http://example.org/gisserver"

    #: Define namespace aliases to use, default is ``{"app": self.xml_namespace}``.
    xml_namespace_aliases = None

    #: Default version to use
    version = "2.0.0"

    #: Allow to set a default service, so the SERVICE parameter can be omitted.
    default_service = None

    #: Supported versions
    accept_versions = ("2.0.0",)

    #: Internal configuration of all available RPC calls.
    accept_operations = {}

    #: Metadata of the service:
    service_description: ServiceDescription | None = None

    #: Template to render an HTML welcome page for non-OGC requests.
    index_template_name = None

    #: Whether to render GET HTML pages
    use_html_templates = True

    #: The OWS request, which contains the complete parsed request.
    ows_request: BaseOwsRequest = None

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        """Render proper XML errors for exceptions on all request types."""
        try:
            return super().dispatch(request, *args, **kwargs)
        except Exception as e:
            response = self.handle_exception(e)
            if response is not None:
                return response
            raise

    def handle_exception(self, exc):
        """Transform an exception into a OGC response.
        When nothing is returned, the exception is raised instead.
        """
        if isinstance(exc, Django_PermissionDenied):
            exc = PermissionDenied(str(exc) or None, locator="typeNames")
            return exc.as_response()
        elif isinstance(exc, OWSException):
            # Wrap our XML-based exception into a response.
            exc.version = self.version  # Use negotiated version
            return exc.as_response()
        else:
            return None

    @classmethod
    def get_xml_namespace_aliases(cls) -> dict[str, str]:
        """Provide all namespaces aliases with a namespace.
        This is most useful for parsing input.
        The default is: ``{"app": cls.xml_namespace}``.
        """
        return cls.xml_namespace_aliases or {"app": cls.xml_namespace}

    @classmethod
    def get_xml_namespaces_to_prefixes(cls) -> dict[str, str]:
        """Provide a mapping from namespace to prefix.
        This is most useful for rendering output.
        The default is: ``{cls.xml_namespace: "app"}``.
        """
        return {
            xml_namespace: prefix
            for prefix, xml_namespace in cls.get_xml_namespace_aliases().items()
        }

    def get(self, request, *args, **kwargs):
        """Entry point to handle HTTP GET requests.

        This parses the 'SERVICE' and 'REQUEST' parameters,
        to call the proper :class:`~gisserver.operations.base.WFSOperation`.

        All query parameters are handled as case-insensitive.
        """
        # Parse GET parameters in Key-Value-Pair syntax format.
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                "Parsing GET parameters:\n%s",
                unquote_plus(
                    request.META["QUERY_STRING"].replace("\n&", "\n").replace("&", "\n")
                ).rstrip(),
            )
        self.kvp = kvp = KVPRequest(request.GET, ns_aliases=self.get_xml_namespace_aliases())

        # Get service (only raises error when value is missing and "default" parameter is not given)
        defaults = {"default": self.default_service} if self.default_service else {}
        service = kvp.get_str("service", **defaults).upper()

        # Perform early version parsing. The detailed validation happens by the operation
        # Parameter objects, but by performing an early check, templates can use that version.
        version = kvp.get_str("version", default=None)
        self.set_version(service, version)

        # Allow for a user-friendly opening page (hence the version check above)
        if self.use_html_templates and self.is_index_request():
            return self.render_index(service)

        # Find the registered operation that handles the request
        operation = kvp.get_str("request")
        wfs_operation_cls = self.get_operation_class(service, operation)

        # Parse the request syntax
        request_cls = wfs_operation_cls.parser_class or resolve_kvp_parser_class(kvp)
        self.ows_request = request_cls.from_kvp_request(kvp)

        # Process the request!
        return self.call_operation(wfs_operation_cls)

    def post(self, request, *args, **kwargs):
        """Entry point to handle HTTP POST requests.

        This parses the XML to get the correct service and operation,
        to call the proper :class:`~gisserver.operations.base.WFSOperation`.
        """
        # Parse the XML body
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Parsing POST:\n%s", request.body.decode().rstrip())
        try:
            root = parse_xml_from_string(
                request.body, extra_ns_aliases=self.get_xml_namespace_aliases()
            )
        except ExternalParsingError as e:
            raise OperationParsingFailed(f"Unable to parse XML: {e}") from e

        # Find the registered operation that handles the request
        service = (
            root.attrib.get("service", self.default_service)
            if self.default_service
            else root.get_str_attribute("service")
        )

        # Perform early version check. version can be omitted for GetCapabilities
        self.set_version(service, root.attrib.get("version"))

        # Find the registered operation that handles the request
        namespace, operation = split_ns(root.tag)
        wfs_operation_cls = self.get_operation_class(service, operation)

        try:
            # Parse the request syntax
            request_cls = wfs_operation_cls.parser_class or resolve_xml_parser_class(root)
            self.ows_request = request_cls.from_xml(root)

            # Process the request!
            return self.call_operation(wfs_operation_cls)
        except XmlElementNotSupported as e:
            # Unknown XML element, e.g. wrong namespace.
            raise OperationNotSupported(str(e), locator=root.attrib.get("handle")) from e
        except (OperationParsingFailed, OperationProcessingFailed) as e:
            # The WFS spec dictates that these exceptions
            # (and ResponseCacheExpired, CannotLockAllFeatures, FeaturesNotLocked which we don't raise)
            # should report the 'handle' in the 'locator' argument of the exception when it was provided.
            if self.ows_request.handle:
                e.locator = self.ows_request.handle
            raise

    def is_index_request(self):
        """Tell whether to index page should be shown."""
        # If none of the typical WFS parameters are given, and there aren't many other
        # (misspelled?) parameters on the query string, this is considered to be an index page.
        # Minimal request has 2 parameters (SERVICE=WFS&REQUEST=GetCapabilities).
        # Some servers also allow a default of SERVICE, thus needing even less.
        return (
            self.request.method == "GET"
            and len(self.request.GET) < 2
            and {
                "REQUEST",
                "SERVICE",
                "VERSION",
                "ACCEPTVERSIONS",
            }.isdisjoint(key.upper() for key in self.request.GET)
        )

    def render_index(self, service: str | None = None):
        """Render the index page."""
        # Not using the whole TemplateResponseMixin config with configurable parameters.
        # If this really needs more configuration, overriding is likely just as easy.
        return render(
            self.request,
            self.get_index_template_names(service),
            self.get_index_context_data(service=service),
        )

    def get_index_context_data(self, service: str | None = None, **kwargs):
        """Provide the context data for the index page."""
        root_url = self.request.build_absolute_uri()

        # Allow passing extra vendor parameters to the links generated in the template
        # (e.g. expand/embed logic)
        base_qs = {
            key: value
            for key, value in self.request.GET.items()
            if key.upper() not in ("SERVICE", "REQUEST", "VERSION", "OUTPUTFORMAT", "TYPENAMES")
        }
        base_query = urlencode(base_qs) + "&" if base_qs else ""

        return {
            "view": self,
            "xml_namespaces": self.get_xml_namespaces_to_prefixes(),
            "service": service,
            "root_url": root_url,
            "base_query": base_query,
            "connect_url": root_url,
            "version": self.version,
            "service_description": self.get_service_description(service),
            "accept_versions": self.accept_versions,
            "accept_operations": self.accept_operations,
            **kwargs,
        }

    def get_index_template_names(self, service: str | None = None):
        """Get the index page template name.
        If no template is configured, some reasonable defaults are selected.
        """
        service = service.lower() if service and service in self.accept_operations else "default"
        if self.index_template_name:
            # Allow the same substitutions for a manually configured template.
            return [self.index_template_name.format(service=service, version=self.version)]
        else:
            return [
                f"gisserver/{service}/{self.version}/index.html",
                f"gisserver/{service}/index.html",
                "gisserver/index.html",
            ]

    def get_operation_class(self, service: str, request: str) -> type[base.WFSOperation]:
        """Resolve the method that the client wants to call."""
        if not self.accept_operations:
            raise ImproperlyConfigured("View has no operations")

        try:
            operations = self.accept_operations[service.upper()]
        except KeyError:
            allowed = ", ".join(sorted(self.accept_operations.keys()))
            raise InvalidParameterValue(
                f"'{service}' is not supported, available are: {allowed}.",
                locator="service",
            ) from None

        # Resolve the operation
        # In mapserver, the operation name is case-insensitive.
        uc_methods = {name.upper(): method for name, method in operations.items()}
        try:
            return uc_methods[request.upper()]
        except KeyError:
            allowed = ", ".join(operations.keys())
            raise OperationNotSupported(
                f"'{request}' is not implemented, supported are: {allowed}.",
                locator="request",
            ) from None

    def call_operation(self, wfs_operation_cls: type[base.WFSOperation]):
        """Call the resolved method."""
        self.request.ows_request = self.ows_request  # for check_permissions()

        wfs_operation = wfs_operation_cls(self, self.ows_request)
        wfs_operation.validate_request(self.ows_request)
        return wfs_operation.process_request(self.ows_request)

    def get_service_description(self, service: str | None = None) -> ServiceDescription:
        """Provide the (dynamically generated) service description."""
        return self.service_description or ServiceDescription(title="Unnamed")

    def set_version(self, service: str, version: str | None):
        """Enforce a particular version based on the request."""
        if service.upper() not in self.accept_operations:
            allowed = ", ".join(sorted(self.accept_operations.keys()))
            raise InvalidParameterValue(
                f"'{service}' is not supported, available are: {allowed}.",
                locator="service",
            ) from None

        if not version:
            return

        if not SAFE_VERSION.match(version):
            # Make really sure we didn't mess things up, as this causes file includes.
            raise SuspiciousOperation("Invalid/insecure version number parsed")

        if version not in self.accept_versions:
            raise InvalidParameterValue(
                f"This server does not support {service} version {version}.", locator="version"
            )

        # Enforce the requested version
        self.version = version

    @property
    def server_url(self):
        """Expose the server URLs for all operations to read."""
        return self.request.build_absolute_uri(self.request.path)


class WFSView(OWSView):
    """A view for a single WFS server.

    This view exposes multiple dataset,
    each containing a single Django model (mapped to feature types in WFS).

    This class can be used by subclassing it, and redefining ``feature_types``
    or ``get_feature_types()``.
    """

    #: Maximum number of features to return
    max_page_size = conf.GISSERVER_DEFAULT_MAX_PAGE_SIZE

    #: Define the features (=tables) in this dataset.
    #: For dynamic per-request logic, consider overwriting :meth:`get_feature_types` instead.
    feature_types: list[FeatureType] = []

    #: Internal configuration of all available RPC calls.
    accept_operations = {
        "WFS": {
            "GetCapabilities": wfs20.GetCapabilities,
            "DescribeFeatureType": wfs20.DescribeFeatureType,
            "GetFeature": wfs20.GetFeature,
            "GetPropertyValue": wfs20.GetPropertyValue,
            "ListStoredQueries": wfs20.ListStoredQueries,
            "DescribeStoredQueries": wfs20.DescribeStoredQueries,
        }
    }

    #: Since URLs to this view are already specifically for WFS,
    #: allow to omit the service name.
    default_service = "WFS"

    #: Metadata of the capabilities:
    wfs_service_constraints = {
        "ImplementsBasicWFS": True,  # Advanced queries/xpath
        "ImplementsTransactionalWFS": False,  # only reads
        "ImplementsLockingWFS": False,  # only basic WFS
        "KVPEncoding": True,  # HTTP GET support
        "XMLEncoding": True,  # HTTP POST requests
        "SOAPEncoding": False,  # no SOAP requests
        "ImplementsInheritance": False,
        "ImplementsRemoteResolve": False,
        "ImplementsResultPaging": True,  # missing next/previous for START / COUNT
        "ImplementsStandardJoins": False,  # returns records as wfs:Tuple in GetFeature
        "ImplementsSpatialJoins": False,
        "ImplementsTemporalJoins": False,
        "ImplementsFeatureVersioning": False,
        "ManageStoredQueries": False,
        #
        # Mentioned as operation constraint in WFS spec:
        "PagingIsTransactionSafe": False,
    }

    #: Metadata of filtering capabilities
    wfs_filter_capabilities = {
        "ImplementsQuery": True,  # <fes:AbstractQueryElement> needed for WFS simple
        "ImplementsAdHocQuery": True,  # <fes:AbstractAdhocQueryElement> needed for WFS simple
        "ImplementsFunctions": True,  # <fes:Function> support
        "ImplementsResourceId": True,  # <fes:ResourceId> support
        "ImplementsMinStandardFilter": True,  # <fes:PropertyIs...> support
        "ImplementsStandardFilter": True,  # <fes:And>, <fes:Or> and advanced functions
        "ImplementsMinSpatialFilter": True,  # <fes:BBOX>
        "ImplementsSpatialFilter": True,  # Other spatial functions
        "ImplementsMinTemporalFilter": False,  # mapserver: true (During)
        "ImplementsTemporalFilter": False,
        "ImplementsVersionNav": False,  # <fes:ResourceId version="..">
        "ImplementsSorting": True,  # SORTBY parameter
        "ImplementsExtendedOperators": False,
        "ImplementsMinimumXPath": True,  # incomplete like mapserver, needed for cite
        "ImplementsSchemaElementFunc": False,
    }

    def get_feature_types(self) -> list[FeatureType]:
        """Return all available feature types this server exposes.

        This method may be overwritten to provide feature types dynamically,
        for example to give them different elements based on user permissions.
        """
        return self.feature_types

    def get_bound_feature_types(self) -> list[FeatureType]:
        """Internal logic wrapping the FeatureType definitions provided by the developer.
        This binds the XML namespace information from this view to the declared types.
        """
        feature_types = self.get_feature_types()
        for feature_type in feature_types:
            # Make sure the feature type can advertise itself with an XML namespace.
            feature_type.bind_namespace(default_xml_namespace=self.xml_namespace)
        return feature_types

    def get_index_context_data(self, **kwargs) -> dict:
        """Get the context data for the index template"""
        context = super().get_index_context_data(**kwargs)
        context.update(
            {
                "GISSERVER_SUPPORTED_CRS_ONLY": conf.GISSERVER_SUPPORTED_CRS_ONLY,
                "wfs_features": self.get_bound_feature_types(),
                "wfs_output_formats": self._get_wfs_output_formats(),
                "wfs_filter_capabilities": self.wfs_filter_capabilities,
                "wfs_service_constraints": self.wfs_service_constraints,
            }
        )
        return context

    def _get_wfs_output_formats(self) -> list[base.OutputFormat]:
        """Find the output formats of the ``GetFeature`` operation for the HTML index."""
        get_feature_operation = self.accept_operations["WFS"]["GetFeature"]
        operation = get_feature_operation(self, ows_request=None)

        # Get output formats, remove duplicates (e.g. GeoJSON/geojson alias)
        wfs_output_formats = []
        seen = set()
        for output_format in operation.get_output_formats():
            if output_format.identifier not in seen:
                wfs_output_formats.append(output_format)
            seen.add(output_format.identifier)
        return wfs_output_formats

    def get_xml_schema_url(self, feature_types: list[FeatureType]) -> str:
        """Return the XML schema URL for the given feature types.
        This is used in the GML output rendering.
        """
        app_namespaces = self.get_xml_namespaces_to_prefixes()
        type_names = ",".join(
            f"{app_namespaces[ft.xml_namespace]}:{ft.name}" for ft in feature_types
        )
        return (
            f"{self.server_url}?SERVICE=WFS&VERSION=2.0.0"
            f"&REQUEST=DescribeFeatureType&TYPENAMES={type_names}"
        )

    def check_permissions(self, feature_type: FeatureType):
        """Hook that allows subclasses to reject access for datasets.
        It may raise a Django PermissionDenied error.

        This can access: ``self.request`` (the Django HTTPRequest) and ``self.ows_request``.
        The latter contains a parsed request object
        from the :mod:`gisserver.parsers.wfs20` package,
        such as the parsed :class:`~gisserver.parsers.wfs20.GetFeature`
        or :class:`~gisserver.parsers.wfs20.GetPropertyValue` request.
        """



"""These classes map to the FES 2.0 specification for expressions.
The class names are identical to those in the FES spec.

Inheritance structure:

* :class:`Expression`

 * :class:`Function` for ``<fes:Function>``.
 * :class:`Literal` for ``<fes:Literal>``.
 * :class:`ValueReference` for ``<fes:ValueReference>``.
 * :class:`BinaryOperator` for FES 1.0 compatibility.

The :class:`BinaryOperator` is included as expression
to handle FES 1.0 :class:`BinaryOperatorType` arithmetic tags:
``<fes:Add>``, ``<fes:Sub>``, ``<fes:Mul>``, ``<fes:Div>``.
"""

from __future__ import annotations

import operator
from dataclasses import dataclass, field
from datetime import date, datetime
from decimal import Decimal as D
from functools import cached_property
from typing import Union

from django.contrib.gis import geos
from django.contrib.gis.db import models as gis_models
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q, Value
from django.db.models.expressions import Combinable

from gisserver.exceptions import ExternalParsingError
from gisserver.extensions.functions import function_registry
from gisserver.parsers.ast import (
    AstNode,
    TagNameEnum,
    expect_no_children,
    expect_tag,
    tag_registry,
)
from gisserver.parsers.gml import (
    GM_Envelope,
    GM_Object,
    TM_Object,
    is_gml_element,
    parse_gml_node,
)
from gisserver.parsers.query import RhsTypes
from gisserver.parsers.values import auto_cast
from gisserver.parsers.xml import NSElement, parse_qname, xmlns
from gisserver.types import XPathMatch, XsdTypes

NoneType = type(None)
ParsedValue = Union[int, str, date, D, datetime, GM_Object, GM_Envelope, TM_Object, NoneType]

OUTPUT_FIELDS = {
    bool: models.BooleanField(),
    str: models.CharField(),
    int: models.IntegerField(),
    date: models.DateField(),
    datetime: models.DateTimeField(),
    float: models.FloatField(),
    D: models.DecimalField(),
    geos.GEOSGeometry: gis_models.GeometryField(),
    geos.Point: gis_models.PointField(),
    geos.LineString: gis_models.LineStringField(),
    geos.LinearRing: gis_models.LineStringField(),
    geos.Polygon: gis_models.PolygonField(),
    geos.MultiPoint: gis_models.MultiPointField(),
    geos.MultiPolygon: gis_models.MultiPolygonField(),
    geos.MultiLineString: gis_models.MultiLineStringField(),
    geos.GeometryCollection: gis_models.GeometryCollectionField(),
}


class BinaryOperatorType(TagNameEnum):
    """FES 1.0 Arithmetic operators.

    These are no longer part of the FES 2.0 spec, but clients (like QGis)
    still assume the server supports these. Hence, these need to be included.
    """

    Add = operator.add
    Sub = operator.sub
    Mul = operator.mul
    Div = operator.truediv


class Expression(AstNode):
    """Abstract base class, as defined by FES spec.

    The FES spec defines the following subclasses:

    * :class:`ValueReference` (pointing to a field name)
    * :class:`Literal` (a scalar value)
    * :class:`Function` (a transformation for a value/field)

    When code uses ``Expression.child_from_xml(element)``, the AST logic will
    initialize the correct subclass for those elements.
    """

    xml_ns = xmlns.fes20

    def build_lhs(self, compiler) -> str:
        """Get the expression as the left-hand-side of the equation.

        This typically returns the expression as a 'field name' which can be
        used in the Django QuerySet.filter(name=...) syntax. When the
        expression is actually a Function/Literal, this should generate the
        name using a queryset annotation.
        """
        value = _make_combinable(self.build_rhs(compiler))
        return compiler.add_annotation(value)

    def build_rhs(self, compiler) -> RhsTypes:
        """Get the expression as the right-hand-side of the equation.

        Typically, this can return the exact value.
        """
        raise NotImplementedError(f"{self.__class__.__name__}.build_rhs()")


@dataclass(repr=False)
@tag_registry.register("Literal")
class Literal(Expression):
    """The ``<fes:Literal>`` element that holds a literal value.

    This can be a string value, possibly annotated with a type::

        <fes:Literal type="xs:boolean">true</fes:Literal>

    Following the spec, the value may also contain a complete geometry::

        <fes:Literal>
            <gml:Envelope xmlns:gml="http://www.opengis.net/gml/3.2" srsName="urn:ogc:def:crs:EPSG::4326">
                <gml:lowerCorner>5.7 53.1</gml:lowerCorner>
                <gml:upperCorner>6.1 53.5</gml:upperCorner>
            </gml:Envelope>
        </fes:Literal>
    """

    # The XSD definition even defines a sequence of xsd:any as possible member!
    raw_value: NoneType | str | GM_Object | GM_Envelope | TM_Object
    raw_type: str | None = None

    def __str__(self):
        return self.value

    def __repr__(self):
        return f"Literal({self.raw_value!r}, raw_type={self.raw_type!r}, type={self.type!r})"

    @cached_property
    def value(self) -> ParsedValue:  # officially <xsd:any>
        """Access the value of the element, cast to the appropriate data type.
        :raises ExternalParsingError: When the value can't be converted to the proper type.
        """
        if not isinstance(self.raw_value, str):
            return self.raw_value  # GML element or None
        elif self.type:
            # Cast the value based on the given xsd:QName
            return self.type.to_python(self.raw_value)
        else:
            # Make sure gt / datetime comparisons work out of the box.
            return auto_cast(self.raw_value)

    @cached_property
    def type(self) -> XsdTypes | None:
        """Tell which datatype the literal holds.
        This returns the type="..." value of the element.
        """
        if not self.raw_type:
            return None

        # The raw type is already translated into a fully qualified XML name,
        # which also allows defining types as "ns0:string", "xs:string" or "xsd:string".
        # These are directly matched to the XsdTypes enum value.
        return XsdTypes(self.raw_type)

    @classmethod
    @expect_tag(xmlns.fes20, "Literal")
    def from_xml(cls, element: NSElement):
        children = len(element)
        if not children:
            # Common case: value is raw text
            raw_value = element.text
        elif children == 1 and is_gml_element(element[0]):
            # Possible: a <gml:Envelope> element to compare something against.
            raw_value = parse_gml_node(element[0])
        else:
            raise ExternalParsingError(
                f"Unsupported child element for <Literal> element: {element[0].tag}."
            )

        return cls(
            raw_value=raw_value,
            raw_type=parse_qname(element.attrib.get("type"), element.ns_aliases),
        )

    def build_lhs(self, compiler) -> str:
        """Alias the value when it's used in the left-hand-side.

        By aliasing the value using an annotation,
        it can be queried like a regular field name.
        """
        return compiler.add_annotation(Value(self.value, output_field=self.get_output_field()))

    def get_output_field(self):
        # When the value is used a left-hand-side, Django needs to know the output type.
        return OUTPUT_FIELDS.get(type(self.value))

    def build_rhs(self, compiler) -> Combinable | Q | ParsedValue:
        """Return the value when it's used in the right-hand-side"""
        return self.value

    def bind_type(self, type: XsdTypes):
        """Assign the expected type that this literal is compared against"""
        if not self.raw_type:
            self.__dict__["type"] = type
            self.__dict__.pop("value", None)  # reset cached_property


@dataclass(repr=False)
@tag_registry.register("ValueReference")
@tag_registry.register("PropertyName", hidden=True)  # FES 1.0 name that old clients still use.
class ValueReference(Expression):
    """The ``<fes:ValueReference>`` element that holds an XPath string.
    In the fes XSD, this is declared as a subclass of xsd:string.

    This parses the syntax like::

        <fes:ValueReference>field-name</fes:ValueReference>
        <fes:ValueReference>path/to/field-name</fes:ValueReference>
        <fes:ValueReference>collection[@attr=value]/field-name</fes:ValueReference>

    The old WFS1/FES1 "PropertyName" is allowed as an alias.
    Various clients still send this, and mapserver/geoserver support this.
    """

    #: The XPath value
    xpath: str
    #: The known namespaces aliases at this point in the XML tree
    xpath_ns_aliases: dict[str, str] | None = field(compare=False, default=None)

    def __str__(self):
        return self.xpath

    def __repr__(self):
        return f"ValueReference({self.xpath!r})"

    @classmethod
    @expect_tag(xmlns.fes20, "ValueReference", "PropertyName")
    @expect_no_children
    def from_xml(cls, element: NSElement):
        return cls(xpath=element.text, xpath_ns_aliases=element.ns_aliases)

    def build_lhs(self, compiler) -> str:
        """Use the field name in a left-hand-side expression."""
        # Optimized from base class: fields don't need an alias lookup through annotations.
        match = self.parse_xpath(compiler.feature_types)
        return match.build_lhs(compiler)

    def build_rhs(self, compiler) -> RhsTypes:
        """Use the field name in a right-hand expression.
        This generates an F-expression for the ORM."""
        match = self.parse_xpath(compiler.feature_types)
        return match.build_rhs(compiler)

    def parse_xpath(self, feature_types: list) -> XPathMatch:
        """Convert the XPath into the required ORM query elements."""
        return feature_types[0].resolve_element(self.xpath, self.xpath_ns_aliases)


@dataclass
@tag_registry.register("Function")
class Function(Expression):
    """The ``<fes:Function name="...">`` element.

    This parses the syntax such as::

        <fes:Function name="Add">
            <fes:ValueReference>field-name</fes:ValueReference>
            <fes:Literal>2</fes:Literal>
        </fes:Function>

    Each argument of the function can be another :class:`Expression`,
    such as a :class:`Function`, :class:`ValueReference` or :class:`Literal`.
    """

    name: str  # scoped name
    arguments: list[Expression]  # xsd:element ref="fes20:expression"

    @classmethod
    @expect_tag(xmlns.fes20, "Function")
    def from_xml(cls, element: NSElement):
        return cls(
            name=element.get_str_attribute("name"),
            arguments=[Expression.child_from_xml(child) for child in element],
        )

    def build_rhs(self, compiler) -> models.Func:
        """Build the SQL function object"""
        db_function = function_registry.resolve_function(self.name)
        args = [arg.build_rhs(compiler) for arg in self.arguments]
        return db_function.build_query(*args)


@dataclass
@tag_registry.register(BinaryOperatorType)
class BinaryOperator(Expression):
    """Support for FES 1.0 arithmetic operators.

    This parses a syntax like::

        <fes:Add>
            <fes:ValueReference>field-name</fes:ValueReference>
            <fes:Literal>2</fes:Literal>
        </fes:Add>

    The operator can be a ``<fes:Add>``, ``<fes:Sub>``, ``<fes:Mul>``, ``<fes:Div>``.

    These are no longer part of the FES 2.0 spec, but clients (like QGis)
    still assume the server use these. Hence, these need to be included.
    """

    _operatorType: BinaryOperatorType
    expression: tuple[Expression, Expression]

    @classmethod
    def from_xml(cls, element: NSElement):
        return cls(
            _operatorType=BinaryOperatorType.from_xml(element),
            expression=(
                Expression.child_from_xml(element[0]),
                Expression.child_from_xml(element[1]),
            ),
        )

    def build_rhs(self, compiler) -> RhsTypes:
        self.validate_arithmetic(compiler, self.expression[0], self.expression[1])
        value1 = self.expression[0].build_rhs(compiler)
        value2 = self.expression[1].build_rhs(compiler)

        # Func() + 2 or 2 + Func() are automatically handled,
        # but Q(..) + 2 or 2 + Q(..) aren't.
        # When both are values, a direct calculation takes place in Python.
        if isinstance(value1, Q):
            value2 = _make_combinable(value2)
        if isinstance(value2, Q):
            value1 = _make_combinable(value1)

        try:
            # This calls somthing like: operator.add(F("field"), 2) or operator.add(1, 2)
            return self._operatorType.value(value1, value2)
        except (TypeError, ValueError, ArithmeticError) as e:
            raise ValidationError(
                f"Invalid data for the 'fes:{self._operatorType.name}' element: {value1} {value2}"
            ) from e

    def validate_arithmetic(self, compiler, lhs: Expression, rhs: Expression):
        """Check whether values support arithmetic operators."""
        if isinstance(lhs, Literal) and isinstance(rhs, ValueReference):
            lhs, rhs = rhs, lhs

        if isinstance(lhs, ValueReference):
            xsd_element = lhs.parse_xpath(compiler.feature_types).child
            if isinstance(rhs, Literal):
                # Since the element is resolved, inform the Literal how to parse the value.
                # This avoids various validation errors along the path.
                rhs.bind_type(xsd_element.type)

                # Validate the expressions against each other
                # This raises an ValidationError when values can't be converted
                xsd_element.to_python(rhs.raw_value)

        return None


def _make_combinable(value) -> Combinable | Q:
    """Make sure the scalar value is wrapped inside a compilable object"""
    if isinstance(value, (Combinable, Q)):
        return value
    else:
        # e.g. str, or GEOSGeometry
        return Value(value, output_field=OUTPUT_FIELDS.get(type(value)))



"""Utilities for building an Abstract Syntax Tree (AST) from an XML fragment.

By transforming the XML Element nodes into Python objects, most logic naturally follows.
For example, the FES filter syntax can be processed into objects that build an ORM query.

Python classes can inherit :class:`AstNode` and register themselves as the parser/handler
for a given tag. Both normal Python classes and dataclass work,
as long as it has an :meth:`AstNode.from_xml` class method.
The custom ``from_xml()`` method should copy the XML data into local attributes.

Next, when :meth:`TagRegistry.node_from_xml` is called,
it will detect which class the XML Element refers to and initialize it using the ``from_xml()`` call.
As convenience, calling a :meth:`AstNode.child_from_xml`
on a subclass will also initialize the right subclass and initialize it.

Since clients may not follow the desired XML schema, and make mistakes, we should guard against
creating an invalid Abstract Syntax Tree. When using :meth:`TagRegistry.node_from_xml`,
the allowed child types can also be provided, preventing invalid child elements.
Furthermore, to support the creation of ``from_xml()`` methods, the :func:`expect_tag`,
:func:`expect_children` and :func:`expect_no_children` decorators validate
whether the given tag has the expected elements. This combination should make it easy
to validate whether a provided XML structure confirms to the supported schema.
"""

from __future__ import annotations

from collections.abc import Iterable
from enum import Enum
from functools import wraps
from typing import TypeVar
from xml.etree.ElementTree import QName

from django.utils.functional import classproperty

from gisserver.exceptions import InvalidXmlElement, XmlElementNotSupported
from gisserver.parsers.xml import NSElement, xmlns

__all__ = (
    "TagNameEnum",
    "AstNode",
    "TagRegistry",
    "tag_registry",
    "expect_children",
    "expect_tag",
    "expect_no_children",
)


class TagNameEnum(Enum):
    """An base clas for enumerations of XML tag names.

    All enumerations that represent tag names inherit from this.
    Each member name should be exactly the XML tag that it refers to.
    """

    @classmethod
    def from_xml(cls, element: NSElement):
        """Cast the element tag name into the enum member.

        This translates the element name
        such as ``{http://www.opengis.net/fes/2.0}PropertyIsEqualTo``
        into a ``PropertyIsEqualTo`` member.
        """
        tag_name = element.tag
        if tag_name.startswith("{"):
            # Split the element tag into the namespace and local name.
            # The stdlib etree doesn't have the properties for this (lxml does).
            end = tag_name.index("}")
            tag_name = tag_name[end + 1 :]

        return cls[tag_name]

    @classmethod
    def _missing_(cls, value):
        raise NotImplementedError(f"<{value}> is not registered as valid {cls.__name__}")

    def __repr__(self):
        # Make repr(filter) easier to copy-paste
        return f"{self.__class__.__name__}.{self.name}"


class AstNode:
    """The base node for all classes that represent an XML tag.

    All subclasses of this class build an Abstract Syntax Tree (AST)
    that describes the XML content in Python objects. Each object can handle
    implement additional logic to

    Each subclass should implement the :meth:`from_xml` to translate
    an XML tag into a Python (data) class.
    """

    #: Default namespace of the element and subclasses, if not given by ``@tag_registry.register()``.
    xml_ns: xmlns | str | None = None

    _xml_tags = []

    @classproperty
    def xml_name(cls) -> str:
        """Tell the default tag by which this class is registered"""
        return cls._xml_tags[0]

    xml_name.__doc__ = "Tell the default tag by which this class is registered"

    def __init_subclass__(cls):
        # Each class level has a fresh list of supported child tags.
        cls._xml_tags = []

    @classmethod
    def from_xml(cls, element: NSElement):
        """Initialize this Python class from the data of the corresponding XML tag.
        Each subclass overrides this to implement the XML parsing of that particular XML tag.
        """
        raise NotImplementedError(
            f"{cls.__name__}.from_xml() is not implemented to parse <{element.tag}>"
        )

    @classmethod
    def child_from_xml(cls, element: NSElement) -> AstNode:
        """Parse the element, returning the correct subclass of this tag.

        When ``Expression.child_from_xml(some_node)`` is given, it may
        return a ``Literal``, ``ValueReference``, ``Function`` or ``BinaryOperator`` node.
        """
        sub_class = tag_registry.resolve_class(element, allowed_types=(cls,))
        return sub_class.from_xml(element)

    @classmethod
    def get_tag_names(cls) -> list[str]:
        """Provide all known XML tags that this code can parse."""
        try:
            # Because a cached class property is hard to build
            return _KNOWN_TAG_NAMES[cls]
        except KeyError:
            all_xml_tags = cls._xml_tags.copy()
            for sub_cls in cls.__subclasses__():
                all_xml_tags.extend(sub_cls.get_tag_names())
            _KNOWN_TAG_NAMES[cls] = all_xml_tags
            return all_xml_tags


_KNOWN_TAG_NAMES = {}

BaseNode = AstNode  # keep old name around
A = TypeVar("A", bound=AstNode)


class TagRegistry:
    """Registration of all classes that can parse XML nodes.

    The same class can be registered multiple times for different tag names.
    """

    parsers: dict[str, type[AstNode]]

    def __init__(self):
        self.parsers = {}

    def register(
        self,
        tag: str | type[TagNameEnum] | None = None,
        namespace: xmlns | str | None = None,
        hidden: bool = False,
    ):
        """Decorator to register a class as XML element parser.

        Usage:

        .. code-block:: python

            @dataclass
            @tag_registry.register()
            class SomeXmlTag(AstNode):
                xml_ns = FES

                @classmethod
                def from_xml(cls, element: NSElement):
                    return cls(
                        ...
                    )

        Whenever an element of the registered XML name is found,
        the given "SomeXmlTag" will be initialized.

        It's also possible to register tag names using an enum;
        each member name is assumed to be an XML tag name.
        """

        def _dec(node_class: type[AstNode]) -> type[AstNode]:
            if tag is None or isinstance(tag, str):
                # Single tag name for the class.
                self._register_tag_parser(
                    node_class, tag=tag or node_class.__name__, namespace=namespace, hidden=hidden
                )
            elif issubclass(tag, TagNameEnum):
                # Allow tags to be an Enum listing possible tag names.
                # Note using __members__, not _member_names_.
                # The latter will skip aliased items (like BBOX/Within).
                for member_name in tag.__members__:
                    self._register_tag_parser(
                        node_class, tag=member_name, namespace=namespace, hidden=hidden
                    )
            else:
                raise TypeError("tag type incorrect")

            return node_class

        return _dec

    def _register_tag_parser(
        self,
        node_class: type[AstNode],
        tag: str,
        namespace: xmlns | str | None = None,
        hidden: bool = False,
    ):
        """Register a Python (data) class as parser for an XML node."""
        if not issubclass(node_class, AstNode):
            raise TypeError(f"{node_class} must be a subclass of AstNode")

        if namespace is None and node_class.xml_ns is None:
            raise RuntimeError(
                f"{node_class.__name__}.xml_ns should be set, or namespace should be given."
            )

        xml_name = QName((namespace or node_class.xml_ns), tag=tag).text
        if xml_name in self.parsers:
            raise RuntimeError(
                f"Another class is already registered to parse the <{xml_name}> tag."
            )

        self.parsers[xml_name] = node_class  # Track this parser to resolve the tag.
        if not hidden:
            node_class._xml_tags.append(xml_name)  # Allow fetching all names later

    def node_from_xml(self, element: NSElement, allowed_types: tuple[type[A]] | None = None) -> A:
        """Find the ``AstNode`` subclass that corresponds to the given XML element,
        and initialize it with the element. This is a convenience shortcut.
        """
        node_class = self.resolve_class(element, allowed_types)
        return node_class.from_xml(element)

    def resolve_class(
        self, element: NSElement, allowed_types: tuple[type[A]] | None = None
    ) -> type[A]:
        """Find the :class:`AstNode` subclass that corresponds to the given XML element."""
        try:
            node_class = self.parsers[element.tag]
        except KeyError:
            msg = f"Unsupported tag: <{element.qname}>"
            if "{" not in element.tag:
                msg = f"{msg} without an XML namespace"
            if allowed_types:
                allowed = _tag_names_to_text(_get_allowed_tag_names(*allowed_types), element)
                msg = f"{msg}, expected one of: {allowed}."

            raise XmlElementNotSupported(msg) from None

        # Check whether the resolved class is indeed a valid option here.
        if allowed_types is not None and not issubclass(node_class, allowed_types):
            types = ", ".join(c.__name__ for c in allowed_types)
            raise InvalidXmlElement(
                f"Unexpected {node_class.__name__} for <{element.qname}> node, "
                f"expected one of: {types}"
            )

        return node_class

    def get_parser_class(self, xml_qname) -> type[AstNode]:
        """Provide the parser class for a given XML Qualified name."""
        return self.parsers[xml_qname]

    def find_subclasses(self, node_type: type[A]) -> list[type[A]]:
        """Find all registered parsers for a given node."""
        return {
            tag: node_class
            for tag, node_class in self.parsers.items()
            if issubclass(node_class, node_type)
        }


def expect_tag(namespace: xmlns | str, *tag_names: str):
    """Decorator for ``from_xml()`` methods that validate whether a given tag is provided.

    For example:

    .. code-block:: python

        @classmethod
        @expect_tag(xmlns.fes20, "Literal")
        def from_xml(cls, element):
            ...

    This guard is needed when nodes are passed directly to a ``from_xml()`` method.
    """
    valid_tags = {QName(namespace, name).text for name in tag_names}
    expect0 = QName(namespace, tag_names[0]).text

    def _wrapper(func):
        @wraps(func)
        def _expect_tag_decorator(cls, element: NSElement, *args, **kwargs):
            if element.tag not in valid_tags:
                raise InvalidXmlElement(
                    f"{cls.__name__} parser expects an <{_replace_common_ns(expect0, element)}> node,"
                    f" got <{element.qname}>"
                )
            return func(cls, element, *args, **kwargs)

        return _expect_tag_decorator

    return _wrapper


def expect_no_children(from_xml_func):
    """Decorator for ``from_xml()`` methods that validate that the XML tag has no child nodes.

    For example:

    .. code-block:: python

        @classmethod
        @expect_tag(xmlns.fes20, "ResourceId")
        @expect_no_children
        def from_xml(cls, element):
            ...
    """

    @wraps(from_xml_func)
    def _expect_no_children_decorator(cls, element: NSElement, *args, **kwargs):
        if len(element):
            raise InvalidXmlElement(
                f"Element <{element.qname}> does not support child elements,"
                f" found <{element[0].qname}>."
            )

        return from_xml_func(cls, element, *args, **kwargs)

    return _expect_no_children_decorator


def expect_children(  # noqa: C901
    min_child_nodes, *expect_types: str | type[AstNode], silent_allowed: tuple[str] = ()
):
    """Decorator for ``from_xml()`` methods to validate whether an element has the expected children.

    For example:

    .. code-block:: python

        @classmethod
        @expect_children(2, Expression)
        def from_xml(cls, element):
            ...
    """
    # Validate arguments early
    for child_type in expect_types + silent_allowed:
        if isinstance(child_type, str):
            if not child_type.startswith("{"):
                raise ValueError(
                    f"String arguments to @expect_children() should be"
                    f" fully qualified XML namespaces, not {child_type!r}"
                )
        elif not isinstance(child_type, type) or not issubclass(child_type, AstNode):
            raise TypeError(f"Unexpected {child_type!r}")

    def _get_allowed(known_tag_names, element):
        return _tag_names_to_text(sorted(set(known_tag_names) - set(silent_allowed)), element)

    def _wrapper(func):
        @wraps(func)
        def _expect_children_decorator(cls, element: NSElement, *args, **kwargs):
            known_tag_names = _get_allowed_tag_names(*expect_types, *silent_allowed)

            if len(element) < min_child_nodes:
                allowed = _get_allowed(known_tag_names, element)
                raise InvalidXmlElement(
                    f"<{element.qname}> should have {min_child_nodes} child nodes,"
                    f" got only {len(element)}."
                    f" Allowed types are: {allowed}."
                )
            for child in element:
                if child.tag not in known_tag_names:
                    allowed = _get_allowed(known_tag_names, element)
                    raise InvalidXmlElement(
                        f"<{element.qname}> does not support a <{child.qname}> child node."
                        f" Allowed types are: {allowed}."
                    )

            return func(cls, element, *args, **kwargs)

        return _expect_children_decorator

    return _wrapper


def _get_allowed_tag_names(*expect_types: type[AstNode] | str) -> list[str]:
    # Resolve arguments later, as get_tag_names() depends on __subclasses__()
    # which may not be completely known at this point.
    tag_names = []
    for child_type in expect_types:
        if isinstance(child_type, type) and issubclass(child_type, AstNode):
            tag_names.extend(child_type.get_tag_names())
        elif isinstance(child_type, str):
            if not child_type.startswith("{"):
                raise ValueError(
                    f"String arguments should be fully qualified XML namespaces, not {child_type!r}"
                )
            tag_names.append(child_type)
        else:
            raise TypeError(f"Unexpected {child_type!r}")
    return tag_names


def _tag_names_to_text(tag_names: Iterable[str], user_element: NSElement) -> str:
    body = _replace_common_ns(">, <".join(tag_names), user_element)
    return f"<{body}>"


def _replace_common_ns(text: str, user_element: NSElement):
    """In error messages, replace the full XML names with QName/prefixed versions.
    The chosen prefixes reference the tags that the client submitted in their XML body.
    """
    for prefix, ns in user_element.ns_aliases.items():
        text = text.replace(f"{{{ns}}}", f"{prefix}:")
    return text


#: The tag registry to register new parsing classes at.
tag_registry = TagRegistry()



"""Helper classes for Coordinate Reference System translations.

This includes the CRS parsing, coordinate transforms and axis orientation.
"""

from __future__ import annotations

import logging
import re
import typing
from dataclasses import dataclass, field
from functools import cached_property, lru_cache

import pyproj
from django.contrib.gis.gdal import AxisOrder, CoordTransform, OGRGeometry, SpatialReference
from django.contrib.gis.geos import GEOSGeometry

from gisserver import conf
from gisserver.exceptions import ExternalValueError

CRS_URN_REGEX = re.compile(
    r"^urn:(?P<domain>[a-z]+)"
    r":def:crs:(?P<authority>[a-z]+)"
    r":(?P<version>[0-9]+(\.[0-9]+(\.[0-9]+)?)?)?"
    r":(?P<id>[0-9]+|crs84)"
    r"$",
    re.IGNORECASE,
)

AnyGeometry = typing.TypeVar("AnyGeometry", GEOSGeometry, OGRGeometry)

__all__ = [
    "CRS",
    "CRS84",
    "WEB_MERCATOR",
    "WGS84",
]

# Caches to avoid reinitializing WGS84 each time.
_COMMON_CRS_BY_STR = {}
_COMMON_CRS_BY_SRID = {}


logger = logging.getLogger(__name__)


@lru_cache(maxsize=200)  # Using lru-cache to avoid repeated GDAL c-object construction
def _get_spatial_reference(srs_input: str | int, srs_type, axis_order):
    """Construct an GDAL object reference"""
    logger.debug(
        "Constructed GDAL SpatialReference(%r, srs_type=%r, axis_order=%s)",
        srs_input,
        srs_type,
        axis_order,
    )
    return SpatialReference(srs_input, srs_type=srs_type, axis_order=axis_order)


@lru_cache(maxsize=100)
def _get_coord_transform(source: SpatialReference, target: SpatialReference) -> CoordTransform:
    """Get an efficient coordinate transformation object.

    The CoordTransform should be used when performing the same
    coordinate transformation repeatedly on different geometries.

    Using a CoordinateTransform also allows setting the AxisOrder setting
    on both ends. When calling ``GEOSGeometry.transform()``, Django will
    create an internal CoordTransform object internally without setting AxisOrder,
    implicitly setting its source SpatialReference to be 'AxisOrder.TRADITIONAL'.
    """
    return CoordTransform(source, target)


_get_proj_crs_from_string = lru_cache(maxsize=10)(pyproj.CRS.from_string)
_get_proj_crs_from_authority = lru_cache(maxsize=10)(pyproj.CRS.from_authority)


@dataclass(frozen=True, eq=False)
class CRS:
    """
    Represents a CRS (Coordinate Reference System), which preferably follows the URN format
    as specified by `the OGC consortium <http://www.opengeospatial.org/ogcUrnPolicy>`_.
    """

    # CRS logic, based upon https://github.com/wglas85/django-wfs/blob/master/wfs/helpers.py
    # Copyright (c) 2006 Wolfgang Glas - Apache 2.0 licensed
    # Ported to Python 3.6 style.

    #: Either "ogc" or "opengis", whereas "ogc" is highly recommended.
    domain: str

    #: Either "OGC" or "EPSG".
    authority: str

    #: The version of the authorities' SRS registry, which is empty or
    #: contains two or three numeric components separated by dots like "6.9" or "6.11.9".
    #: For WFS 2.0 this is typically empty.
    version: str

    #: A string representation of the coordinate system reference ID.
    #: For OGC, only "CRS84" is supported as crsid. For EPSG, this is the formatted CRSID.
    crsid: str

    #: The integer representing the numeric spatial reference ID as
    #: used by the EPSG and GIS database backends.
    srid: int

    #: GDAL SpatialReference with PROJ.4 / WKT content to describe the exact transformation.
    backends: tuple[SpatialReference, SpatialReference] = (None, None)

    #: Original input
    origin: str = field(init=False, default=None)

    #: Tell whether the input format used the legacy notation.
    force_xy: bool = False

    @classmethod
    def from_string(cls, uri: str | int) -> CRS:
        """
        Parse an CRS (Coordinate Reference System) URI, which preferably follows the URN format
        as specified by `the OGC consortium <http://www.opengeospatial.org/ogcUrnPolicy>`_
        and construct a new CRS instance.

        The value can be 3 things:

        * A URI in OGC URN format.
        * A legacy CRS URI ("epsg:<SRID>", or "http://www.opengis.net/...").
        * A numeric SRID (which calls :meth:`from_srid()`)
        """
        if known_crs := _COMMON_CRS_BY_STR.get(uri):
            return known_crs  # Avoid object re-creation

        if isinstance(uri, int) or uri.isdigit():
            return cls.from_srid(int(uri))
        elif uri.startswith("urn:"):
            return cls._from_urn(uri)
        else:
            return cls._from_prefix(uri)

    @classmethod
    def from_srid(cls, srid: int):
        """Instantiate this class using a numeric spatial reference ID

        This is logically identical to calling::

            CRS.from_string("urn:ogc:def:crs:EPSG::<SRID>")
        """
        if common_crs := _COMMON_CRS_BY_SRID.get(srid):
            return common_crs  # Avoid object re-creation

        return cls(
            domain="ogc",
            authority="EPSG",
            version="",
            crsid=str(srid),
            srid=int(srid),
        )

    @classmethod
    def _from_urn(cls, urn):  # noqa: C901
        """Instantiate this class using a URN format.
        This format is defined in https://portal.ogc.org/files/?artifact_id=30575.
        """
        urn_match = CRS_URN_REGEX.match(urn)
        if not urn_match:
            raise ExternalValueError(f"Unknown CRS URN [{urn}] specified: {CRS_URN_REGEX.pattern}")

        domain = urn_match.group("domain")
        authority = urn_match.group("authority").upper()

        if domain not in ("ogc", "opengis"):
            raise ExternalValueError(f"CRS URI [{urn}] contains unknown domain [{domain}]")

        if authority == "EPSG":
            crsid = urn_match.group("id")
            try:
                srid = int(crsid)
            except ValueError:
                raise ExternalValueError(
                    f"CRS URI [{urn}] should contain a numeric SRID value."
                ) from None
        elif authority == "OGC":
            # urn:ogc:def:crs:OGC::CRS84 has x/y ordering (longitude/latitude)
            crsid = urn_match.group("id").upper()
            if crsid not in ("CRS84", "84"):
                raise ExternalValueError(f"OGC CRS URI from [{urn}] contains unknown id [{id}]")
            srid = 4326
        else:
            raise ExternalValueError(f"CRS URI [{urn}] contains unknown authority [{authority}]")

        crs = cls(
            domain=domain,
            authority=authority,
            version=urn_match.group(3) or "",
            crsid=crsid,
            srid=srid,
        )
        crs.__dict__["origin"] = urn
        return crs

    @classmethod
    def _from_prefix(cls, uri):
        """Instantiate this class from a non-URI notation.

        The modern URL format (:samp:`http://www.opengis.net/def/crs/epsg/0/{xxxx}`)
        is defined in https://portal.ogc.org/files/?artifact_id=46361.

        Older notations like :samp:`EPSG:{xxxx}` or legacy XML URLs like
        and :samp`:http://www.opengis.net/gml/srs/epsg.xml#{xxxx}` are also supported.
        """
        # Make sure origin uses the expected upper/lowercasing.
        origin = uri.lower() if "://" in uri else uri.upper()
        for prefix, force_xy in (
            (
                "EPSG:",
                (conf.GISSERVER_FORCE_XY_EPSG_4326 and origin == "EPSG:4326"),
            ),
            (
                "http://www.opengis.net/gml/srs/epsg.xml#",
                conf.GISSERVER_FORCE_XY_OLD_CRS,
            ),
            ("http://www.opengis.net/def/crs/epsg/0/", False),
        ):
            if origin.startswith(prefix):
                crsid = origin[len(prefix) :]
                try:
                    srid = int(crsid)
                except ValueError:
                    raise ExternalValueError(
                        f"CRS URI [{uri}] should contain a numeric SRID value."
                    ) from None

                crs = cls(
                    domain="ogc",
                    authority="EPSG",
                    version="",
                    crsid=crsid,
                    srid=srid,
                    force_xy=force_xy,
                )
                crs.__dict__["origin"] = origin
                return crs

        raise ExternalValueError(f"Unknown CRS URI [{uri}] specified")

    @property
    def legacy(self):
        """Return a legacy string in the format :samp:`http://www.opengis.net/gml/srs/epsg.xml#{srid}`."""
        # This mirrors what GeoSever does, as this notation always has an axis ordering defined.
        # Notations like EPSG:xxxx notation don't have such consistent usage.
        return f"http://www.opengis.net/gml/srs/epsg.xml#{self.srid:d}"

    @cached_property
    def urn(self):
        """Return The OGC URN corresponding to this CRS."""
        return f"urn:{self.domain}:def:crs:{self.authority}:{self.version or ''}:{self.crsid}"

    @property
    def is_north_east_order(self) -> bool:
        """Tell whether the axis is in north/east ordering."""
        return self.axis_direction == ["north", "east"]

    @cached_property
    def axis_direction(self) -> list[str]:
        """Tell what the axis ordering of this coordinate system is.

        For example, WGS84 will return ``['north', 'east']``.

        While computer systems typically use X,Y, other systems may use northing/easting.
        Historically, latitude was easier to measure and given first.
        In physics, using 'radial, polar, azimuthal' is again a different perspective.
        See: https://wiki.osgeo.org/wiki/Axis_Order_Confusion for a good summary.
        """
        proj_crs = self._as_proj()
        return [axis.direction for axis in proj_crs.axis_info]

    def __str__(self):
        return self.legacy if self.force_xy else self.urn

    def __eq__(self, other):
        if isinstance(other, CRS):
            return self.matches(other, compare_legacy=True)
        else:
            return NotImplemented

    def matches(self, other, compare_legacy=True) -> bool:
        """Tell whether this CRS is identical to another one."""
        return (
            # "urn:ogc:def:crs:EPSG::4326" != "urn:ogc:def:crs:OGC::CRS84"
            # even through they both share the same srid.
            self.srid == other.srid
            and self.authority == other.authority
            and (
                # Also, legacy notations like EPSG:4326 are treated differently,
                # unless this is configured to not make a difference.
                not compare_legacy
                or self.force_xy == other.force_xy
            )
        )

    def __hash__(self):
        """Used to match objects in a set."""
        return hash((self.authority, self.srid))

    def _as_gdal(self, axis_order: AxisOrder) -> SpatialReference:
        """Generate the GDAL Spatial Reference object."""
        if self.backends[axis_order] is None:
            backends = list(self.backends)
            if self.origin and "://" not in self.origin:  # avoid downloads and OGR errors
                # Passing the origin helps to detect CRS84 strings
                backends[axis_order] = _get_spatial_reference(self.origin, "user", axis_order)
            else:
                backends[axis_order] = _get_spatial_reference(self.srid, "epsg", axis_order)

            # Write back in "readonly" format.
            self.__dict__["backends"] = tuple(backends)

        return self.backends[axis_order]

    def _as_proj(self) -> pyproj.CRS:
        """Generate the PROJ CRS object"""
        if (
            isinstance(self.origin, str)
            and not self.origin.isdigit()
            and "epsg.xml#" not in self.origin  # not supported
        ):
            # Passing the origin helps to detect CRS84 strings
            return _get_proj_crs_from_string(self.origin)
        else:
            return _get_proj_crs_from_authority(self.authority, self.srid)

    def apply_to(
        self,
        geometry: AnyGeometry,
        clone=False,
        axis_order: AxisOrder | None = None,
    ) -> AnyGeometry | None:
        """Transform the geometry using this coordinate reference.

        Every transformation within this package happens through this method,
        giving full control over coordinate transformations.

        A bit of background: geometries are provided as ``GEOSGeometry`` from the database.
        This is basically a simple C-based storage implementing "OpenGIS Simple Features for SQL",
        except it does *not* store axis orientation. These are assumed to be x/y.

        To perform transformations, GeoDjango loads the GEOS-geometry into GDAL/OGR.
        The transformed geometry is loaded back to GEOS. To avoid this conversion,
        pass the OGR object directly and continue working on that.

        Internally, this method caches the used GDAL ``CoordTransform`` object,
        so repeated transformations of the same coordinate systems are faster.

        The axis order can change during the transformation.
        From a programming perspective, screen coordinates (x/y) were traditionally used.
        However, various systems and industries have always worked with north/east (y/x).
        This includes systems with critical safety requirements in aviation and maritime.
        The CRS authority reflects this practice. What you need depends on the use case:

        * The (GML) output of WFS 2.0 and WMS 1.3 respect the axis ordering of the CRS.
        * GeoJSON always provides coordinates in x/y, to keep web-based clients simple.
        * PostGIS stores the data in x/y.
        * WFS 1.0 used x/y, WFS 1.3 used y/x except for ``EPSG:4326``.

        When receiving legacy notations (e.g. ``EPSG:4326`` instead of ``urn:ogc:def:crs:EPSG::4326``),
        the data is still projected in legacy ordering, unless ``GISSERVER_FORCE_XY_...`` is disabled.
        This reflects the design of `GeoServer Axis Ordering
        <https://docs.geoserver.org/stable/en/user/services/wfs/axis_order.html>`_
        to have maximum interoperability with legacy/JavaScript clients.

        After GDAL/OGR changed the axis orientation, that information is
        lost when the return value is loaded back into GEOS.
        To address this, :meth:`tag_geometry` is called on the result.

        :param geometry: The GEOS Geometry, or GDAL/OGR loaded geometry.
        :param clone: Whether the object is changed in-place, or a copy is returned.
                      For GEOS->GDAL->GEOS conversions, this makes no difference in efficiency.
        :param axis_order: Which axis ordering to convert the geometry into (depends on the use-case).
        """
        if axis_order is None:
            # This transforms by default to WFS 2 axis ordering (e.g. latitude/longitude),
            # unless a legacy notation is used (e.g. EPSG:4326).
            axis_order = AxisOrder.TRADITIONAL if self.force_xy else AxisOrder.AUTHORITY

        if isinstance(geometry, OGRGeometry):
            transform = _get_coord_transform(geometry.srs, self._as_gdal(axis_order=axis_order))
            return geometry.transform(transform, clone=clone)
        else:
            # See if the geometry was tagged with a CRS.
            # When the data comes from an unknown source, assume this is from database storage.
            # PostGIS stores data in longitude/latitude (x/y) ordering, even for srid 4326.
            # By passing the 'AxisOrder.TRADITIONAL', a conversion from 4326 to 4326
            # with 'AxisOrder.AUTHORITY' will detect that coordinate ordering needs to be changed.
            source_axis_order = getattr(geometry, "_axis_order", AxisOrder.TRADITIONAL)

            if self.srid == geometry.srid and source_axis_order == axis_order:
                # Avoid changes if spatial reference system is identical, and no axis need to change.
                if clone:
                    return geometry.clone()
                else:
                    return None

            # Get GDAL spatial reference for converting coordinates (uses proj internally).
            # Using a cached coordinate transform object (is faster for repeated transforms)
            # The object is also tagged so another apply_to() call would recognize the state.
            source = _get_spatial_reference(geometry.srid, "epsg", source_axis_order)
            target = self._as_gdal(axis_order=axis_order)
            transform = _get_coord_transform(source, target)

            # Transform
            geometry = geometry.transform(transform, clone=clone)
            if clone:
                self.tag_geometry(geometry, axis_order=axis_order)
            return geometry

    @classmethod
    def tag_geometry(self, geometry: GEOSGeometry, axis_order: AxisOrder):
        """Associate this object with the geometry.

        This informs the :meth:`apply_to` method that this source geometry
        already had the correct axis ordering (e.g. it was part of the ``<fes:BBOX>`` logic).
        The srid integer doesn't communicate that information.
        """
        geometry._axis_order = axis_order

    def cache_instance(self):
        """Cache a common CRS, no need to re-instantiate the same object again.
        This also makes sure that requests which use the same URN will get our CRS object
        version, instead of a fresh new one.
        """
        if self.authority == "EPSG" and self.srid != 4326:
            # Only register for EPSG to avoid conflicting axis ordering issues.
            # (WGS84 and CRS84 both use srid 4326, and the 'EPSG:4326' notation is treated as legacy)
            _COMMON_CRS_BY_SRID[self.srid] = self

        _COMMON_CRS_BY_STR[str(self)] = self


#: Worldwide GPS, latitude/longitude (y/x). https://epsg.io/4326
WGS84 = CRS.from_string("urn:ogc:def:crs:EPSG::4326")

#: GeoJSON default. This is like WGS84 but with longitude/latitude (x/y).
CRS84 = CRS.from_string("urn:ogc:def:crs:OGC::CRS84")

#: Spherical Mercator (Google Maps, Bing Maps, OpenStreetMap, ...), see https://epsg.io/3857
WEB_MERCATOR = CRS.from_string("urn:ogc:def:crs:EPSG::3857")


# Register these common ones:
WGS84.cache_instance()
CRS84.cache_instance()
WEB_MERCATOR.cache_instance()



"""Handle adhoc-query objects.

The adhoc query is based on incoming request parameters,
such as the "FILTER", "BBOX" and "RESOURCEID" parameters.

These definitions follow the WFS spec.
"""

from __future__ import annotations

import logging
from dataclasses import dataclass
from functools import cached_property

from django.db.models import Q

from gisserver.crs import CRS
from gisserver.exceptions import (
    InvalidParameterValue,
    MissingParameterValue,
    OperationNotSupported,
)
from gisserver.parsers import fes20
from gisserver.parsers.ast import tag_registry
from gisserver.parsers.ows import KVPRequest
from gisserver.parsers.query import CompiledQuery
from gisserver.parsers.xml import NSElement, xmlns
from gisserver.projection import FeatureProjection

from .base import QueryExpression
from .projection import PropertyName

logger = logging.getLogger(__name__)

ADHOC_QUERY_ELEMENTS = (PropertyName, fes20.Filter, fes20.SortBy)


@dataclass
@tag_registry.register("Query", xmlns.wfs)
class AdhocQuery(QueryExpression):
    """The Ad hoc query expression parameters.

    This parses and handles the syntax::

        <wfs:Query typeNames="...">
            <wfs:PropertyName>...</wfs:PropertyName>
            <fes:Filter>...</fes:Filter>
            <fes:SortBy>...</fes:SortBy>
        </wfs:Query>

    And supports the KVP syntax:

    .. code-block:: urlencoded

        ?SERVICE=WFS&...&TYPENAMES=ns:myType&FILTER=...&SORTBY=...&SRSNAME=...&PROPERTYNAME=...
        ?SERVICE=WFS&...&TYPENAMES=ns:myType&BBOX=...&SORTBY=...&SRSNAME=...&PROPERTYNAME=...
        ?SERVICE=WFS&...&TYPENAMES=ns:myType&RESOURCEID=...

    This represents all dynamic queries received as request (hence "adhoc"),
    such as the "FILTER" and "BBOX" arguments from an HTTP GET.

    The WFS Spec has 3 class levels for this:

     - AdhocQueryExpression (types, projection, selection, sorting)
     - Query (adds srsName, featureVersion)
     - StoredQuery (adds storedQueryID)

    For KVP requests, this class seems almost identifical to the provided parameters.
    However, the KVP format allows to provide parameter lists,
    to perform support multiple queries in a single request!

    .. seealso::
        https://www.mediamaps.ch/ogc/schemas-xsdoc/sld/1.2/query_xsd.html#AbstractAdhocQueryExpressionType
    """

    # Tag attributes (implements ``fes:AbstractAdhocQueryExpression``)
    # WFS allows multiple names to construct JOIN queries.
    # See https://docs.ogc.org/is/09-025r2/09-025r2.html#107
    # and https://docs.ogc.org/is/09-025r2/09-025r2.html#190

    #: The 'typeNames' value if the request provided them. use :meth:`get_type_names` instead.
    typeNames: list[str]
    #: Aliases for typeNames are used for joining the same table twice. (JOIN statements are not supported yet).
    aliases: list[str] | None = None
    #: For XML POST requests, this handle value is returned in the ``<ows:Exception>``.
    handle: str = ""

    # part of the <wfs:Query> tag attributes:
    #: The Coordinate Reference System to render the tag in
    srsName: CRS | None = None

    #: Projection clause (implements ``fes:AbstractProjectionClause``)
    property_names: list[PropertyName] | None = None

    #: Selection clause (implements ``fes:AbstractSelectionClause``).
    #: - for XML POST this is encoded in a <fes:Filter> tag.
    #: - for HTTP GET, this is encoded as FILTER, FILTER_LANGUAGE, RESOURCEID, BBOX.
    filter: fes20.Filter | None = None

    #: Sorting Clause (implements ``fes:AbstractSortingClause``)
    sortBy: fes20.SortBy | None = None

    def __post_init__(self):
        if len(self.typeNames) > 1:
            raise OperationNotSupported("Join queries are not supported", locator="typeNames")
        if self.aliases:
            raise OperationNotSupported("Join queries are not supported", locator="aliases")

    @classmethod
    def from_xml(cls, element: NSElement) -> AdhocQuery:
        """Parse the XML element of the Query tag."""
        type_names = [
            element.parse_qname(qname)
            for qname in element.get_str_attribute("typeNames").split(" ")
        ]
        aliases = element.attrib.get("aliases", None)
        srsName = element.attrib.get("srsName", None)
        property_names = []
        filter = None
        sortBy = None

        for child in element:
            # The FES XSD dictates the element ordering, but this is ignored here.
            node = tag_registry.node_from_xml(child, allowed_types=ADHOC_QUERY_ELEMENTS)
            if isinstance(node, PropertyName):
                property_names.append(node)
            elif isinstance(node, fes20.Filter):
                filter = node
            elif isinstance(node, fes20.SortBy):
                sortBy = node
            else:
                raise NotImplementedError(
                    f"Parsing {node.__class__} not handled in AdhocQuery.from_xml()"
                )

        return AdhocQuery(
            typeNames=type_names,
            aliases=aliases.split(" ") if aliases is not None else None,
            handle=element.attrib.get("handle", ""),
            property_names=property_names or None,
            filter=filter,
            sortBy=sortBy,
            srsName=CRS.from_string(srsName) if srsName else None,
        )

    @classmethod
    def from_kvp_request(cls, kvp: KVPRequest):
        """Build this object from an HTTP GET (key-value-pair) request.

        Note the caller should have split the KVP request parameter lists.
        This class only handles a single parameter pair.
        """
        # Parse attributes
        typeNames = [
            kvp.parse_qname(qname)
            for qname in kvp.get_list("typeNames", alias="TYPENAME", default=[])
        ]
        aliases = kvp.get_list("aliases", default=None)
        srsName = kvp.get_custom("srsName", default=None, parser=CRS.from_string)

        # KVP requests may omit the typenames if RESOURCEID=... is given.
        if not typeNames and "RESOURCEID" not in kvp:
            raise MissingParameterValue("Empty TYPENAMES parameter", locator="typeNames")

        # Parse elements
        filter = fes20.Filter.from_kvp_request(kvp)
        sort_by = fes20.SortBy.from_kvp_request(kvp)

        # Parse projection
        property_names = None
        if "PROPERTYNAME" in kvp:
            names = kvp.get_list("propertyName", default=[])
            # Check for WFS 1.x syntax of ?PROPERTYNAME=*
            if names != ["*"]:
                property_names = [
                    PropertyName(xpath=name, xpath_ns_aliases=kvp.ns_aliases) for name in names
                ]

        return AdhocQuery(
            # Attributes
            typeNames=typeNames,
            aliases=aliases,
            srsName=srsName,
            # Elements
            property_names=property_names,
            filter=filter,
            sortBy=sort_by,
        )

    @cached_property
    def query_locator(self):
        """Overrides the 'query_locator' attribute, so the 'locator' argument is correctly set."""
        if self.filter is not None and self.filter.get_resource_id_types():
            return "resourceId"
        else:
            return "filter"

    def get_type_names(self) -> list[str]:
        """Tell which type names this query uses.
        Multiple values means a JOIN is made (not supported yet).
        """
        if not self.typeNames and self.filter is not None:
            # Also make the behavior consistent, always supply the type name.
            return self.filter.get_resource_id_types() or []
        else:
            return self.typeNames

    def get_projection(self) -> FeatureProjection:
        """Tell how the ``<wfs:Query>`` element should be displayed."""
        return FeatureProjection(
            self.feature_types,
            self.property_names,
            self.value_reference,
            output_crs=self.srsName,
        )

    def bind(self, *args, **kwargs):
        """Override to make sure the 'locator' points to the actual object that defined the type."""
        try:
            super().bind(*args, **kwargs)
        except InvalidParameterValue as e:
            if not self.typeNames:
                e.locator = "resourceId"
            raise

        # Validate the srsName too
        if self.srsName is not None:
            self.srsName = self.feature_types[0].resolve_crs(self.srsName, locator="srsName")

    def build_query(self, compiler: CompiledQuery) -> Q | None:
        """Apply our collected filter data to the compiler."""
        # Add the sorting
        if self.sortBy is not None:
            self.sortBy.build_ordering(compiler)

        if self.filter is not None:
            # Generate the internal query object from the <fes:Filter>,
            # this can return a Q object.
            return self.filter.build_query(compiler)
        else:
            return None

    def as_kvp(self) -> dict:
        """Translate the POST request into KVP GET parameters. This is needed for pagination."""
        params = super().as_kvp()
        params["TYPENAMES"] = ",".join(self.typeNames)
        if self.srsName is not None:
            params["SRSNAME"] = str(self.srsName)

        if self.filter is not None:
            raise NotImplementedError()  # not going to parse that, nor does mapserver
        if self.sortBy is not None:
            params["SORTBY"] = self.sortBy.as_kvp()

        return params



"""Additional parsing logic for GML values.

Most GML elements are parsed through GeoDjango by using the GEOSGeometry element.
"""

from __future__ import annotations

from gisserver.exceptions import ExternalParsingError
from gisserver.parsers.ast import tag_registry
from gisserver.parsers.xml import NSElement, parse_xml_from_string

from .base import AbstractGeometry, GM_Envelope, GM_Object, TM_Object
from .geometries import GEOSGMLGeometry, is_gml_element  # also do tag registration

__all__ = [
    "GM_Object",
    "GM_Envelope",
    "TM_Object",
    "AbstractGeometry",
    "GEOSGMLGeometry",
    "parse_gml",
    "parse_gml_node",
    "find_gml_nodes",
]


def parse_gml(text: str | bytes) -> GM_Object | GM_Envelope | TM_Object:
    """Parse an XML <gml:...> string."""
    root_element = parse_xml_from_string(text)
    return parse_gml_node(root_element)


def parse_gml_node(element: NSElement) -> GM_Object | GM_Envelope | TM_Object:
    """Parse the GML element."""
    if not is_gml_element(element):
        raise ExternalParsingError(f"Expected GML namespace for {element.tag}")

    # All known root nodes as GML object:
    return tag_registry.node_from_xml(element, allowed_types=(GM_Object, GM_Envelope, TM_Object))


def find_gml_nodes(element: NSElement) -> list[NSElement]:
    """Find all ``<gml:...>`` elements in a node.
    This selects all GML elements, including GML 2.1 tags.
    """
    return [child for child in element if is_gml_element(child)]



"""WFS 2.0 element parsing.

These classes parse the XML request body.

The full spec can be found at: https://www.ogc.org/publications/standard/wfs/.
Secondly, using https://www.mediamaps.ch/ogc/schemas-xsdoc/sld/1.2/wfs_xsd.html can be very
helpful to see which options each object type should support.
"""

from .adhoc import AdhocQuery
from .base import QueryExpression
from .projection import PropertyName
from .requests import (
    DescribeFeatureType,
    DescribeStoredQueries,
    GetCapabilities,
    GetFeature,
    GetPropertyValue,
    ListStoredQueries,
    ResultType,
)
from .stored import StoredQuery

__all__ = (
    # Requests
    "GetCapabilities",
    "DescribeFeatureType",
    "GetFeature",
    "GetPropertyValue",
    "ListStoredQueries",
    "DescribeStoredQueries",
    "ResultType",
    # Queries
    "QueryExpression",
    "AdhocQuery",
    "StoredQuery",
    # Projection
    "PropertyName",
)



"""The base classes for GML.

See "Table D.2" in the GML 3.2.1 spec, showing how the UML names
map to the GML implementations. These names are referenced by the FES spec.
"""

from __future__ import annotations

from gisserver.parsers.ast import AstNode
from gisserver.parsers.query import CompiledQuery

__all__ = (
    "GM_Object",
    "GM_Envelope",
    "TM_Object",
    "AbstractTimeObject",
    "AbstractTimePrimitive",
    "Envelope",
    "AbstractGeometry",
)


class GM_Object(AstNode):
    """Abstract base classes for all GML objects, regardless of their version."""

    def build_rhs(self, compiler: CompiledQuery):
        """Required function to implement.
        This allows using the value to be used in a binary operator.
        """
        raise NotImplementedError()


class GM_Envelope(AstNode):
    """Abstract base classes for the GML envelope, regardless of their version."""


class TM_Object(AstNode):
    """Abstract base classes for temporal GML objects, regardless of their version.

    See ISO 19108 TM_Object (see D.2.5.2 and ISO 19108:2002, 5.2.2)
    """


# Instead of polluting the MRO with unneeded base classes, create aliases:

#: The ``<gml:AbstractGeometry>`` definition implements the ISO 19107 GM_Object.
AbstractGeometry = GM_Object

#: The ``<gml:Envelope>`` implements ISO 19107 GM_Envelope (see D.2.3.4 and ISO 19107:2003, 6.4.3).
Envelope = GM_Envelope

# See https://www.mediamaps.ch/ogc/schemas-xsdoc/sld/1.1.0/temporal_xsd.html

#: The base class for all time objects
AbstractTimeObject = TM_Object

#: The base classes for time primitives.
AbstractTimePrimitive = AbstractTimeObject



"""Exceptions for the WFS operations.

Not all exception codes are listed here, only the one that
apply to the current conformance class.

See:
https://docs.opengeospatial.org/is/09-025r2/09-025r2.html#35
https://docs.opengeospatial.org/is/09-025r2/09-025r2.html#411
"""

from __future__ import annotations

import logging
import typing
from contextlib import contextmanager

from django.conf import settings
from django.core.exceptions import FieldError, ValidationError
from django.db import InternalError, ProgrammingError
from django.http import HttpResponse
from django.utils.html import format_html

from gisserver import conf

if typing.TYPE_CHECKING:
    from gisserver.parsers import wfs20

logger = logging.getLogger(__name__)


@contextmanager
def wrap_parser_errors(name: str, locator: str):
    """Convert the value into a Python format.
    This catches any typical exceptions and transforms them into an OWSException.
    """
    try:
        yield
    except ExternalParsingError as e:
        raise OperationParsingFailed(
            f"Unable to parse {name} argument: {e}", locator=locator
        ) from None
    except (TypeError, ValueError, NotImplementedError) as e:
        # TypeError/ValueError are raised by most handlers for unexpected data
        # The NotImplementedError can be raised by fes parsing.
        raise InvalidParameterValue(f"Invalid {name} argument: {e}", locator=locator) from None


@contextmanager
def wrap_filter_errors(query: wfs20.QueryExpression):  # noqa:C901
    """Perform a QuerySet/filter creation operation.
    and trap many parser errors in the making of it."""
    try:
        yield
    except ExternalParsingError as e:
        # Bad input data
        _log_filter_error(query, logging.ERROR, e)
        raise OperationParsingFailed(str(e), locator=query.query_locator) from e
    except ExternalValueError as e:
        # Bad input data
        _log_filter_error(query, logging.ERROR, e)
        raise InvalidParameterValue(str(e), locator=query.query_locator) from e
    except ValidationError as e:
        # Bad input data
        _log_filter_error(query, logging.ERROR, e)
        raise OperationParsingFailed(
            "\n".join(map(str, e.messages)),
            locator=query.query_locator,
        ) from e
    except FieldError as e:
        # e.g. doing a LIKE on a foreign key, or requesting an unknown field.
        if not conf.GISSERVER_WRAP_FILTER_DB_ERRORS:
            raise
        _log_filter_error(query, logging.ERROR, e)
        raise InvalidParameterValue(
            "Internal error when processing filter",
            locator=query.query_locator,
        ) from e
    except (InternalError, ProgrammingError) as e:
        # e.g. comparing datetime against integer
        if not conf.GISSERVER_WRAP_FILTER_DB_ERRORS:
            raise
        logger.exception("WFS request failed: %s\nQuery: %r", str(e), query)
        msg = str(e)
        locator = "srsName" if "Cannot find SRID" in msg else query.query_locator
        raise InvalidParameterValue(f"Invalid request: {msg}", locator=locator) from e
    except (TypeError, ValueError) as e:
        # TypeError/ValueError could reference a datatype mismatch in an
        # ORM query, but it could also be an internal bug. In most cases,
        # this is already caught by XsdElement.validate_comparison().
        raise
        if _is_orm_error(e):
            if query.query_locator == "STOREDQUERY_ID":
                # This is a fallback, ideally the stored query performs its own validation.
                raise InvalidParameterValue(
                    f"Invalid stored query parameter: {e}", locator=query.query_locator
                ) from e
            else:
                raise InvalidParameterValue(
                    f"Invalid filter query: {e}", locator=query.query_locator
                ) from e
        raise


def _is_orm_error(exception: Exception) -> bool:
    """Tell whether an exception is caused by the ORM."""
    traceback = exception.__traceback__
    while traceback.tb_next is not None:
        traceback = traceback.tb_next
        if "/django/db/models/query" in traceback.tb_frame.f_code.co_filename:
            return True
    return False


def _log_filter_error(query, level, exc):
    """Report a filtering parsing error in the logging"""
    filter = getattr(query, "filter", None)  # AdhocQuery only
    fes_xml = filter.source if filter is not None else "(not provided)"
    try:
        sql = exc.__cause__.cursor.query.decode()
    except AttributeError:
        logger.log(level, "WFS query failed: %s\nFilter:\n%s", exc, fes_xml)
    else:
        logger.log(
            level,
            "WFS query failed: %s\nSQL Query: %s\n\nFilter:\n%s",
            exc,
            sql,
            fes_xml,
        )


class ExternalValueError(ValueError):
    """Raise a ValueError for external input.
    This helps to distinguish between internal bugs
    (e.g. unpacking values) and malformed external input.
    """


class ExternalParsingError(ValueError):
    """Raise a ValueError for a parsing problem."""


class XmlElementNotSupported(ExternalParsingError):
    """Raise a ValueError when an XML tag is not known by the parser at all."""


class InvalidXmlElement(ExternalParsingError):
    """Raise a ValueError when a particular XML tag wasn't expected."""


class OWSException(Exception):
    """Base class for XML based exceptions in this module."""

    status_code = 400  # Most common code in spec
    reason = None
    service = None
    version = "2.0.0"
    code = None
    text_template = None
    debug_hint = True

    def __init__(self, text=None, code=None, locator=None, status_code=None):
        text = text or self.text_template.format(code=self.code, locator=locator)
        if (code and len(text) < len(code)) or (locator and len(text) < len(locator)):
            raise ValueError(f"text/locator arguments are switched: {text!r}, locator={locator!r}")

        super().__init__(text)
        self.locator = locator
        self.text = text
        self.code = code or self.code or self.__class__.__name__
        self.status_code = status_code or self.status_code

    def as_response(self) -> HttpResponse:
        """Return the excetion as HTTP response."""
        logger.debug("Returning HTTP %d for %s: %s", self.status_code, self.code, self.text)
        xml_body = self.as_xml()
        return HttpResponse(
            b'<?xml version="1.0" encoding="UTF-8"?>\n%b' % xml_body.encode("utf-8"),
            content_type="text/xml; charset=utf-8",
            status=self.status_code,
            reason=self.reason,
        )

    def as_xml(self) -> str:
        """Serialize the exception to an XML string."""
        return format_html(
            "<ows:ExceptionReport"
            ' xmlns:ows="http://www.opengis.net/ows/1.1"'
            ' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"'
            ' xsi:schemaLocation="http://www.opengis.net/ows/1.1'
            ' http://schemas.opengis.net/ows/1.1.0/owsExceptionReport.xsd"'
            ' xml:lang="en-US"'
            ' version="2.0.0">\n'
            '  <ows:Exception exceptionCode="{code}"{locator_attr}>\n\n'
            "    <ows:ExceptionText>{text}{debug}</ows:ExceptionText>\n\n"
            "  </ows:Exception>\n"
            "</ows:ExceptionReport>\n",
            code=self.code,
            locator_attr=(
                format_html(' locator="{locator}"', locator=self.locator) if self.locator else ""
            ),
            text=self.text,
            debug=(
                ".\n\n(set GISSERVER_WRAP_FILTER_DB_ERRORS=False to see the Django error page)"
                if settings.DEBUG and self.status_code >= 500 and self.debug_hint
                else ""
            ),
        )

    def __html__(self):
        return self.as_xml()


class WFSException(OWSException):
    service = "WFS"


class OperationNotSupported(WFSException):
    """WFS Method is called, but does not exist on this server."""

    status_code = 400
    reason = "Not Implemented"
    code = "OperationNotSupported"
    text_template = "Operation is not implemented."


class OperationParsingFailed(WFSException):
    """Error parsing the request to a WFS method."""

    status_code = 400
    code = "OperationParsingFailed"
    text_template = "The request could not be parsed by the server."


class OperationProcessingFailed(WFSException):
    """Error while processing the request."""

    status_code = 500  # 400 & 403 may also be used but are deprecated.
    reason = "Server processing failed"
    code = "OperationProcessingFailed"
    text_template = "The request could not be processed by the server."


class MissingParameterValue(WFSException):
    """Required parameter is missing"""

    status_code = 400
    code = "MissingParameterValue"
    text_template = "Missing required '{locator}' parameter."


class InvalidParameterValue(WFSException):
    """Unsupported choice, e.g. unsupported CRS, bad resultType."""

    status_code = 400
    code = "InvalidParameterValue"
    text_template = "Invalid value for '{locator}' parameter."


class VersionNegotiationFailed(WFSException):
    """GetCapabilities called with unsupported versions."""

    status_code = 400
    code = "VersionNegotiationFailed"
    text_template = "'ACCEPTVERSIONS' contains an invalid version number."


class NotFound(WFSException):
    """The requested ResourceId could not be found."""

    status_code = 404
    reason = "Invalid feature or property value"
    code = "NotFound"
    text_template = "Invalid feature or property value"


class OptionNotSupported(WFSException):
    status_code = 400
    reason = "Not Implemented"
    code = "OptionNotSupported"
    text_template = "Option is not implemented."


class PermissionDenied(WFSException):
    """Permission denied (custom error).

    Note this error is not part of the spec,
    but it's still useful to have access controls.
    """

    status_code = 403
    reason = "Permission Denied"
    code = "PermissionDenied"
    text_template = "You do not have permission to perform this action."



"""The output rendering classes for all response types.

This also includes collection classes for iterating over the Django QuerySet results.
"""

from .results import FeatureCollection, SimpleFeatureCollection  # isort: skip (fixes import loops)

from .base import CollectionOutputRenderer, OutputRenderer, XmlOutputRenderer
from .csv import CSVRenderer, DBCSVRenderer
from .geojson import DBGeoJsonRenderer, GeoJsonRenderer
from .gml32 import (
    DBGML32Renderer,
    DBGML32ValueRenderer,
    GML32Renderer,
    GML32ValueRenderer,
)
from .stored import DescribeStoredQueriesRenderer, ListStoredQueriesRenderer
from .utils import to_qname
from .xmlschema import XmlSchemaRenderer

__all__ = [
    "OutputRenderer",
    "CollectionOutputRenderer",
    "FeatureCollection",
    "SimpleFeatureCollection",
    "DBCSVRenderer",
    "DBGeoJsonRenderer",
    "DBGML32Renderer",
    "DBGML32ValueRenderer",
    "CSVRenderer",
    "GeoJsonRenderer",
    "GML32Renderer",
    "GML32ValueRenderer",
    "XmlOutputRenderer",
    "XmlSchemaRenderer",
    "ListStoredQueriesRenderer",
    "DescribeStoredQueriesRenderer",
    "to_qname",
]



"""General utilities for outputting XML content"""

from datetime import date, datetime, time, timezone
from decimal import Decimal as D

from django.core.exceptions import ImproperlyConfigured

from gisserver.parsers.xml import xmlns

AUTO_STR = (int, float, D, date, time)

COMMON_NAMESPACES = xmlns.as_namespaces()


__all__ = (
    "attr_escape",
    "tag_escape",
    "to_qname",
    "render_xmlns_attributes",
    "value_to_text",
    "value_to_xml_string",
)


def tag_escape(s: str):
    """Escape a value for usage in XML text."""
    return s.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")


def attr_escape(s: str):
    """Escape a value for usage in an XML attribute.
    This is slightly faster than ``html.escape()`` as it doesn't replace single quotes.
    """
    # Having tried all possible variants, this code still outperforms other forms of escaping.
    return s.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;").replace('"', "&quot;")


def value_to_xml_string(value):
    """Format a Python value for usage in XML text."""
    # Simple scalar value
    if isinstance(value, str):  # most cases
        return tag_escape(value)
    elif isinstance(value, datetime):
        return value.astimezone(timezone.utc).isoformat()
    elif isinstance(value, bool):
        return "true" if value else "false"
    elif isinstance(value, AUTO_STR):
        return value  # no need for _tag_escape(), and f"{value}" works faster.
    else:
        return tag_escape(str(value))


def value_to_text(value):
    """Format a Python value for usage in plain text output.
    This doesn't do any XML escaping.
    """
    # Simple scalar value, no XML escapes
    if isinstance(value, str):  # most cases
        return value
    elif isinstance(value, datetime):
        return value.astimezone(timezone.utc).isoformat()
    elif isinstance(value, bool):
        return "true" if value else "false"
    else:
        return value  # f"{value} works faster and produces the right format.


def render_xmlns_attributes(xml_namespaces: dict[str, str]):
    """Render XML Namespace declaration attributes, i.e. ``xmlns:prefix="uri"`` for each dict item."""
    return " ".join(
        f'xmlns:{prefix}="{xml_namespace}"' if prefix else f'xmlns="{xml_namespace}"'
        for xml_namespace, prefix in xml_namespaces.items()
    )


def to_qname(namespace, localname, namespaces: dict[str, str]) -> str:
    """Convert a fully qualified XML tag name to a prefixed short name."""
    if namespace is None:
        return localname

    prefix = namespaces.get(namespace)  # allow ""
    if prefix is None:
        try:
            prefix = COMMON_NAMESPACES[namespace]
        except KeyError:
            raise ImproperlyConfigured(
                f"No XML namespace prefix defined for '{namespace}'.\n"
                "This can be configured in 'WFSView.xml_namespace_aliases'."
            ) from None

    return f"{prefix}:{localname}" if prefix else localname



"""Handle (stored)query objects.

These definitions follow the WFS spec.
"""

from __future__ import annotations

import typing
from dataclasses import dataclass, field
from functools import cached_property
from typing import Any, ClassVar

from gisserver.exceptions import (
    ExternalParsingError,
    InvalidParameterValue,
    MissingParameterValue,
)
from gisserver.extensions.queries import (
    StoredQueryDescription,
    StoredQueryImplementation,
    stored_query_registry,
)
from gisserver.parsers.ast import tag_registry
from gisserver.parsers.ows import KVPRequest
from gisserver.parsers.query import CompiledQuery
from gisserver.parsers.xml import NSElement, xmlns
from gisserver.projection import FeatureProjection

from .base import QueryExpression

if typing.TYPE_CHECKING:
    from gisserver.output import SimpleFeatureCollection

__all__ = ("StoredQuery",)

# Fully qualified tag names
WFS_PARAMETER = xmlns.wfs20.qname("Parameter")


@dataclass
@tag_registry.register("StoredQuery", xmlns.wfs)
class StoredQuery(QueryExpression):
    """The ``<wfs:StoredQuery>`` element.

    This loads a predefined query on the server.
    A good description can be found at:
    https://mapserver.org/ogc/wfs_server.html#stored-queries-wfs-2-0

    This parses the following syntax::

        <wfs:StoredQuery handle="" id="">
          <wfs:Parameter name="">...</wfs:Parameter>
          <wfs:Parameter name="">...</wfs:Parameter>
        </wfs:StoredQuery>

    and the KVP syntax::

        ?SERVICE=WFS&REQUEST=GetFeature&VERSION=2.0.0&STOREDQUERY_ID=...&{parameter}=...

    Note that the base class logic (such as ``<wfs:PropertyName>`` elements) are still applicable.

    This element resolves the stored query using the
    :class:`~gisserver.extensions.queries.StoredQueryRegistry`,
    and passes the execution to this custom function.
    """

    query_locator: ClassVar[str] = "STOREDQUERY_ID"

    id: str
    parameters: dict[str, Any]
    ns_aliases: dict[str, str] = field(compare=False, default_factory=dict)

    @classmethod
    def from_kvp_request(cls, kvp: KVPRequest):
        """Parse the KVP request syntax. Any query parameters are additional parameters at the query string."""
        query_id = kvp.get_str("STOREDQUERY_ID")
        query_description = stored_query_registry.resolve_query(query_id)
        raw_values = {
            # Take the value when it's there. No validation is done yet,
            # as those error messages are nicer in _parse_parameters()
            name: kvp.get_str(name)
            for name in query_description.parameters
            if name in kvp
        }

        # Avoid unexpected behavior, check whether the client also sends adhoc query parameters
        uc_args = [name.upper() for name in raw_values]
        for name in ("filter", "bbox", "resourceID"):
            uc_name = name.upper()
            if uc_name not in uc_args and uc_name in kvp:
                raise InvalidParameterValue(
                    "Stored query can't be combined with adhoc-query parameters", locator=name
                )

        return cls(
            id=query_id,
            parameters=cls._parse_parameters(query_description, raw_values),
            ns_aliases=kvp.ns_aliases,
        )

    @classmethod
    def from_xml(cls, element: NSElement):
        """Read the XML element."""
        query_id = element.get_str_attribute("id")
        query_description = stored_query_registry.resolve_query(query_id)
        raw_values = {
            parameter.get_str_attribute("name"): parameter.text
            for parameter in element.findall(WFS_PARAMETER)
        }
        return cls(
            id=query_id,
            parameters=cls._parse_parameters(query_description, raw_values),
            ns_aliases=element.ns_aliases,
        )

    @classmethod
    def _parse_parameters(
        cls, query_description: StoredQueryDescription, raw_values: dict[str, str]
    ) -> dict[str, Any]:
        """Validate and translate the incoming parameter.
        This transforms the raw values into their Python types.
        It also validates whether ll expected parameters are provided.
        """
        values = {}
        for name, xsd_type in query_description.parameters.items():
            try:
                raw_value = raw_values.pop(name)
            except KeyError:
                raise MissingParameterValue(
                    f"Stored query {query_description.id} requires an '{name}' parameter",
                    locator=name,
                ) from None

            try:
                values[name] = xsd_type.to_python(raw_value)
            except ExternalParsingError as e:
                raise InvalidParameterValue(
                    f"Stored query {query_description.id} parameter '{name}' can't parse '{raw_value}' as {xsd_type}."
                ) from e

        # Anything left means more parameters were given
        if raw_values:
            names = ", ".join(raw_values)
            raise InvalidParameterValue(
                f"Stored query {query_description.id} does not support the parameter: '{names}'.",
                locator=next(iter(raw_values)),
            )

        return values

    @cached_property
    def implementation(self) -> StoredQueryImplementation:
        """Initialize the stored query from this request."""
        query_description = stored_query_registry.resolve_query(self.id)
        return query_description.implementation_class(
            **self.parameters, ns_aliases=self.ns_aliases
        )

    def bind(self, *args, **kwargs):
        """Associate this query with the application context."""
        super().bind(*args, **kwargs)
        self.implementation.bind(source_query=self, feature_types=self.feature_types)

    def build_query(self, compiler: CompiledQuery):
        """Forward queryset creation to the implementation class."""
        return self.implementation.build_query(compiler)

    def as_kvp(self) -> dict:
        """Translate the POST request into KVP GET parameters. This is needed for pagination."""
        # As this is such edge case, only support the minimal for CITE tests.
        params = super().as_kvp()
        params["STOREDQUERY_ID"] = self.id
        for name, value in self.parameters.items():
            params[name] = str(value)  # should be raw value, but good enough for now.
        return params

    def get_type_names(self) -> list[str]:
        """Tell which features are touched by the query."""
        return self.implementation.get_type_names()

    def get_projection(self) -> FeatureProjection:
        """Tell how the <wfs:StoredQuery> output should be displayed."""
        return FeatureProjection(
            self.feature_types,
            self.property_names,
            value_reference=self.value_reference,
            # In the spec, it's not possible to change the output CRS of a stored query:
            # output_crs=self.srsName,
            output_standalone=self.implementation.has_standalone_output,  # for GetFeatureById
        )

    def finalize_results(self, result: SimpleFeatureCollection):
        return self.implementation.finalize_results(result)



from __future__ import annotations

from urllib.parse import urljoin

from django.template import Library

from gisserver.features import FeatureType
from gisserver.output import to_qname
from gisserver.types import XsdAnyType, XsdNode

register = Library()


@register.filter(name="urljoin")
def urljoin_(fragment, server_url):
    return urljoin(server_url, fragment)


@register.filter(name="to_qname")
def _to_qname(xsd_type: XsdNode | XsdAnyType, xml_namespaces=None):
    """Translate a full XML name to a shortened name, using common prefixes.."""
    return to_qname(xsd_type.namespace, xsd_type.name, xml_namespaces or {})


@register.simple_tag(name="feature_qname", takes_context=True)
def feature_qname(context, feature_type: FeatureType, xml_namespaces=None):
    """Translate a full XML name to a shortened name."""
    if xml_namespaces is None:
        xml_namespaces = context["xml_namespaces"]
    return to_qname(feature_type.xml_namespace, feature_type.name, xml_namespaces)



"""Helper classes to handle geometry data types.

The bounding box can be calculated within Python, or read from a database result.
"""

from __future__ import annotations

import math
from dataclasses import dataclass

from django.contrib.gis.geos import GEOSGeometry

from gisserver.crs import CRS, CRS84, WEB_MERCATOR, WGS84  # noqa: F401 (keep old exports)

#: The CRS for the ``<ows:WGS84BoundingBox>`` element:
WGS84_BOUNDING_BOX_CRS = CRS.from_string("urn:ogc:def:crs:OGC:2:84")

__all__ = [
    "BoundingBox",
    "WGS84BoundingBox",
]


@dataclass
class BoundingBox:
    """A bounding box.
    Due to the overlap between 2 types, this element is used for 2 cases:

    * The ``<ows:WGS84BoundingBox>`` element for ``GetCapabilities``.
    * The ``<gml:Envelope>`` inside an ``<gml:boundedBy>``  single feature.

    While both classes have no common base class (and exist in different schema's),
    their properties are identical.

    The X/Y coordinates can be either latitude or longitude, depending on the CRS.

    Note this isn't using the GDAL/OGR "Envelope" object, as that doesn't expose the CRS,
    and requires constant copies to merge geometries.
    """

    min_x: float
    min_y: float
    max_x: float
    max_y: float
    crs: CRS | None = None

    @classmethod
    def from_geometries(cls, geometries: list[GEOSGeometry], crs: CRS) -> BoundingBox | None:
        """Calculate the extent of a collection of geometries."""
        if not geometries:
            return None
        elif len(geometries) == 1:
            # Common case: feature has a single geometry
            ogr_geometry = geometries[0].ogr
            crs.apply_to(ogr_geometry, clone=False)
            return cls(*ogr_geometry.extent, crs=crs)
        else:
            # Feature has multiple geometries.
            # Start with an obviously invalid bbox,
            # which corrects at the first extend_to_geometry call.
            result = cls(math.inf, math.inf, -math.inf, -math.inf, crs=crs)

            for geometry in geometries:
                ogr_geometry = geometry.ogr
                crs.apply_to(ogr_geometry, clone=False)
                result.extend_to(*ogr_geometry.extent)

            return result

    @property
    def lower_corner(self):
        return [self.min_x, self.min_y]

    @property
    def upper_corner(self):
        return [self.max_x, self.max_y]

    def extend_to(self, min_x: float, min_y: float, max_x: float, max_y: float):
        """Expand the bounding box in-place"""
        self.min_x = min(self.min_x, min_x)
        self.min_y = min(self.min_y, min_y)
        self.max_x = max(self.max_x, max_x)
        self.max_y = max(self.max_y, max_y)

    def __add__(self, other):
        """Combine both extents into a larger box."""
        if isinstance(other, BoundingBox):
            if other.crs != self.crs:
                raise ValueError(
                    "Can't combine instances with different spatial reference systems"
                )
            return self.__class__(
                min(self.min_x, other.min_x),
                min(self.min_y, other.min_y),
                max(self.max_x, other.max_x),
                max(self.max_y, other.max_y),
                crs=self.crs,
            )
        else:
            return NotImplemented


class WGS84BoundingBox(BoundingBox):
    """The ``<ows:WGS84BoundingBox>`` element for the ``GetCapabilities`` element.

    This always has coordinates are always in longitude/latitude axis ordering,
    the CRS is fixed to ``urn:ogc:def:crs:OGC:2:84``.
    """

    def __init__(self, min_x: float, min_y: float, max_x: float, max_y: float):
        super().__init__(min_x, min_y, max_x, max_y, crs=WGS84_BOUNDING_BOX_CRS)



from __future__ import annotations

import typing
from collections import deque
from collections.abc import Iterable
from io import StringIO
from typing import cast

from gisserver.features import FeatureType
from gisserver.parsers.xml import xmlns
from gisserver.types import XsdComplexType, XsdElement

from .base import XmlOutputRenderer
from .utils import to_qname

if typing.TYPE_CHECKING:
    from gisserver.operations.base import WFSOperation


class XmlSchemaRenderer(XmlOutputRenderer):
    """Output rendering for DescribeFeatureType.

    This renders a valid XSD schema that describes the data type of the feature.
    """

    content_type = "application/gml+xml; version=3.2"  # mandatory for WFS
    xml_namespaces = {
        # Default extra namespaces to include in the xmlns="..." attributes
        "http://www.w3.org/2001/XMLSchema": "",  # no xs:string, but "string"
        "http://www.opengis.net/gml/3.2": "gml",
    }

    def __init__(self, operation: WFSOperation, feature_types: list[FeatureType]):
        """Overwritten method to handle the DescribeFeatureType context."""
        super().__init__(operation)
        self.feature_types = feature_types

        # For rendering type="..." fields, avoid xs: prefix.
        self.type_namespaces = self.app_namespaces.copy()
        self.type_namespaces[xmlns.xs.value] = ""  # no "xs:string" but "string"

    def get_headers(self):
        """Make wget output slightly nicer."""
        typenames = "+".join(feature_type.name for feature_type in self.feature_types)
        return {"Content-Disposition": f'inline; filename="{typenames}.xsd"'}

    def render_stream(self):
        # Render first with original app_namespaces
        xmlns_attrib = self.render_xmlns_attributes()

        # Allow targetNamespace to be unprefixed by self.node_to_qname().
        target_namespace = self.feature_types[0].xml_namespace
        self.app_namespaces[target_namespace] = ""  # no "app:element" but "element"

        self.output = output = StringIO()
        output.write(
            f"""<?xml version='1.0' encoding="UTF-8" ?>
<schema {xmlns_attrib}
   targetNamespace="{target_namespace}"
   elementFormDefault="qualified" version="0.1">

"""
        )
        output.write(self.render_imports())
        output.write("\n")

        for feature_type in self.feature_types:
            self.write_feature_type(feature_type)

        output.write("</schema>\n")
        return output.getvalue()

    def render_imports(self):
        return (
            '  <import namespace="http://www.opengis.net/gml/3.2"'
            ' schemaLocation="http://schemas.opengis.net/gml/3.2.1/gml.xsd" />\n'
        )

    def write_feature_type(self, feature_type: FeatureType):
        xsd_type: XsdComplexType = feature_type.xsd_type

        # This declares the that a top-level <app:featureName> is a class of a type.
        xsd_type_qname = self.to_qname(xsd_type, self.type_namespaces)
        feature_qname = to_qname(
            feature_type.xml_namespace, feature_type.name, self.app_namespaces
        )
        self.output.write(
            f'  <element name="{feature_qname}"'
            f' type="{xsd_type_qname}" substitutionGroup="gml:AbstractFeature" />\n\n'
        )

        # Next, the complexType is rendered that defines the element contents.
        # Next, the complexType(s) are rendered that defines the element contents.
        # In case any fields are expanded (hence become subtypes), these are also included.
        self.write_complex_type(xsd_type)
        for complex_type in self._get_complex_types(xsd_type):
            self.write_complex_type(complex_type)

    def write_complex_type(self, complex_type: XsdComplexType):
        """Write the definition of a single class."""
        complex_qname = self.to_qname(complex_type)
        self.output.write(f'  <complexType name="{complex_qname}">\n')

        if complex_type.base is not None:
            base_qname = self.to_qname(complex_type.base)
            self.output.write(
                f"    <complexContent>\n"  # extend base class
                f'      <extension base="{base_qname}">\n'
            )
            indent = "   "
        else:
            indent = ""

        self.output.write(f"    {indent}<sequence>\n")

        for xsd_element in complex_type.elements:
            self.output.write(f"      {indent}{self.render_element(xsd_element)}\n")

        self.output.write(f"    {indent}</sequence>\n")
        if complex_type.base is not None:
            self.output.write(
                "      </extension>\n"  # end extension
                "    </complexContent>\n"
            )
        self.output.write("  </complexType>\n\n")

    def render_element(self, xsd_element: XsdElement):
        """Staticmethod for unit testing."""
        qname = self.to_qname(xsd_element)
        type_qname = self.to_qname(xsd_element.type, self.type_namespaces)

        attributes = [f'name="{qname}" type="{type_qname}"']
        if xsd_element.min_occurs is not None:
            attributes.append(f'minOccurs="{xsd_element.min_occurs}"')
        if xsd_element.max_occurs is not None:
            attributes.append(f'maxOccurs="{xsd_element.max_occurs}"')
        if xsd_element.nillable:
            str_bool = "true" if xsd_element.nillable else "false"
            attributes.append(f'nillable="{str_bool}"')

        return "<element {} />".format(" ".join(attributes))

    def _get_complex_types(self, root: XsdComplexType) -> Iterable[XsdComplexType]:
        """Find all fields that reference to complex types, including nested elements."""
        elements = deque(root.elements)
        complex_types = {}

        # Walk through next types, unless they are already seen.
        # No recursion is used here to handle circular loops of types.
        while elements:
            element = elements.popleft()
            element_type = element.type
            if element_type.is_complex_type and element_type.name not in complex_types:
                # ComplexType was not seen before, register it.
                # It's members are added to the deque for recursive processing.
                xsd_element_type = cast(XsdComplexType, element_type)
                complex_types[element_type.name] = xsd_element_type
                elements.extend(xsd_element_type.elements)

        # Present in a consistent order
        return complex_types.values()



"""XML parsing for all incoming requests.

This logic uses the etree logic from the standard library,
with some extra extensions to expose the original namespace aliases.
Using defusedxml, incoming DOS attacks are prevented.

To handle more complex XML structures, consider building an Abstract Syntax Tree (AST)
to translate the XML Element classes into Python objects.
The :mod:`gisserver.parsers.ast` module provides the building blocks for that.
"""

from __future__ import annotations

import logging
import typing
from enum import Enum
from xml.etree.ElementTree import Element, QName, TreeBuilder

from defusedxml.ElementTree import DefusedXMLParser, ParseError

from gisserver.exceptions import ExternalParsingError, wrap_parser_errors

logger = logging.getLogger(__name__)

__all__ = (
    "xmlns",
    "NSElement",
    "parse_xml_from_string",
    "parse_qname",
    "split_ns",
)


class xmlns(Enum):
    """Common namespaces within WFS land.
    Note these short aliases are arbitrary in XML syntax; the XML code may use any alias (such as ns0).
    The full qualified name (e.g. ``<{http://www.opengis.net/gml/3.2}Point>``) is the actual tag name.
    """

    # XML standard
    xml = "http://www.w3.org/XML/1998/namespace"
    xsd = "http://www.w3.org/2001/XMLSchema"
    xsi = "http://www.w3.org/2001/XMLSchema-instance"
    xlink = "http://www.w3.org/1999/xlink"

    # APIs by the Open Geospatial Consortium (OGC)
    ogc = "http://www.opengis.net/ogc"
    ows10 = "http://www.opengis.net/ows"  # OGC Web Service (OWS) base classes
    ows11 = "http://www.opengis.net/ows/1.1"
    ows20 = "http://www.opengis.net/ows/2.0"
    wms = "http://www.opengis.net/wms"  # Web Map Service (WMS)
    wcs = "http://www.opengis.net/wcs"  # Web Coverage Service (WCS)
    wcs20 = "http://www.opengis.net/wcs/2.0"
    wps = "http://www.opengis.net/wps/1.0.0"  # Web Processing Service (WPS)
    wmts = "http://www.opengis.net/wmts/1.0"  # Web Map Tile Service (WMTS)
    wfs1 = "http://www.opengis.net/wfs"
    wfs20 = "http://www.opengis.net/wfs/2.0"  # Web Feature Service (WFS)
    fes20 = "http://www.opengis.net/fes/2.0"  # Filter Encoding Standard (FES)
    gml21 = "http://www.opengis.net/gml"
    gml32 = "http://www.opengis.net/gml/3.2"

    # Internal aliases
    ows = ows11  # alias to currently used version (WFS 2.0 uses OWS 1.1)
    wfs = wfs20
    fes = fes20
    gml = gml32  # alias to latest version
    xs = xsd  # commonly used

    @classmethod
    def as_ns_aliases(cls) -> dict[str, str]:
        """Map the namespaces as {alias: uri}"""
        return {prefix: member.value for prefix, member in cls.__members__.items()}

    @classmethod
    def as_namespaces(cls) -> dict[str, str]:
        """Map the namespaces as {uri: alias}. This will use the common aliases (without version numbers)"""
        return {member.value: prefix for prefix, member in cls._member_map_.items()}

    def __str__(self):
        # Python 3.11+ has StrEnum for this.
        return self.value

    def qname(self, local_name) -> str:
        """Convert the tag name into a fully qualified name."""
        return f"{{{self.value}}}{local_name}"  # same as QName(..).text

    def __contains__(self, tag: NSElement | str) -> bool:
        """Tell whether a given tag exists in this namespace"""
        if isinstance(tag, NSElement):
            tag = tag.tag
        elif not isinstance(tag, str):
            return False
        return tag.startswith(f"{{{self.value}}}")


class NSElement(Element):
    """Custom XML element, which also exposes its original namespace aliases.
    That information is needed to parse text content and attributes in WFS
    that hold a QName value. For example:

    * ``<ValueReference>ns0:elementName</ValueReference>``
    * ``<Query typeNames="ns1:name">``
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ns_aliases = {}  # assigned by NSTreeBuilder, in {prefix: uri} format.

    def parse_qname(self, qname: str) -> str:
        """Resolve an aliased QName value to its fully qualified name."""
        return parse_qname(qname, self.ns_aliases)

    @property
    def qname(self) -> str:
        """Provde the tag name in its original short format"""
        ns, localname = split_ns(self.tag)
        if ns:
            for prefix, full_ns in self.ns_aliases.items():
                if full_ns == ns:
                    return f"{prefix}:{localname}" if prefix else localname
        return localname

    def get_str_attribute(self, name: str) -> str:
        """Resolve an attribute, raise an error when it's missing."""
        try:
            return self.attrib[name]
        except KeyError:
            raise ExternalParsingError(
                f"Element {self.tag} misses required attribute '{name}'"
            ) from None

    def get_int_attribute(self, name: str, default=None) -> int | None:
        """Retrieve the integer value from an element attribute."""
        value = self.attrib.get(name)
        if value is None:
            return default

        with wrap_parser_errors(name, locator=name):
            return int(value)

    if typing.TYPE_CHECKING:
        # Make sure the type checking knows the actual type of the elements.
        def find(self, path: str, namespaces: dict[str, str] | None = None) -> NSElement | None:
            return super().find(path, namespaces)

        def findall(self, path: str, namespaces: dict[str, str] | None = None) -> list[NSElement]:
            return super().findall(path, namespaces)

        def __iter__(self) -> typing.Iterator[NSElement]:
            return super().__iter__()


def parse_qname(qname: str | None, ns_aliases: dict) -> str | None:
    """Resolve the QName aliases.

    For example, ``gml:Point`` will be resolved to ``{http://www.opengis.net/gml/3.2}Point``.
    The XML namespace prefix is a custom alias, so if "ns0" is declared as "http://www.opengis.net/gml/3.2",
    it means "ns0:Point" should resolve to the same fully qualified type name.
    """
    if not qname:
        return None

    if "/" in qname:
        raise ExternalParsingError(f"Can't resolve QName '{qname}', this is an XPath notation.")

    # Allow resolving @gml:id, remove the @ sigm.
    is_attribute = qname[0] == "@"
    if is_attribute:
        qname = qname[1:]

    prefix, _, localname = qname.rpartition(":")
    if not prefix and ("" not in ns_aliases or ns_aliases[""] == xmlns.wfs20.value):
        # In case a request uses <GetFeature xmlns="http://www.opengis.net/wfs/2.0">,
        # non-prefixed QName values will be interpreted as existing in "wfs" namespace. Avoid that.
        full_name = localname
    else:
        try:
            uri = ns_aliases[prefix]
        except KeyError:
            logger.debug("Can't resolve QName '%s', available namespaces: %r", qname, ns_aliases)
            raise ExternalParsingError(
                f"Can't resolve QName '{qname}', an XML namespace declaration is missing."
            ) from None

        full_name = QName(uri, localname).text

    return f"@{full_name}" if is_attribute else full_name


class NSTreeBuilder(TreeBuilder):
    """Custom TreeBuilder to track namespaces."""

    def __init__(self, extra_ns_aliases, **kwargs):
        super().__init__(element_factory=NSElement, **kwargs)
        # A new stack level is added directly, as start_ns() is called before start()
        if extra_ns_aliases:
            self.ns_stack = [extra_ns_aliases, {}]
        else:
            self.ns_stack = [{}]

    def start(self, tag, attrs):
        super().start(tag, attrs)
        self.ns_stack.append({})  # reserve stack for child tags

    def start_ns(self, prefix, uri):
        self.ns_stack[-1][prefix] = uri

    def end(self, tag) -> Element:
        element = super().end(tag)
        self.ns_stack.pop()  # clear reservation for child tags
        element.ns_aliases = self._flatten_ns()
        return element

    def _flatten_ns(self) -> dict:
        result = {}
        for level in self.ns_stack:
            result.update(level)
        return result


def parse_xml_from_string(
    xml_string: str | bytes, extra_ns_aliases: dict[str, str] | None = None
) -> NSElement:
    """Provide a safe and consistent way for parsing XML.

    This uses a custom parser, so namespace aliases can be tracked.
    All elements also have an :attr:`ns_aliases` attribute that exposes
    the original alias that was used for the namespace.
    """
    # Passing a custom parser potentially circumvents defusedxml,
    # so note the parser is again configured in the same way:
    parser = DefusedXMLParser(
        target=NSTreeBuilder(extra_ns_aliases=extra_ns_aliases),
        forbid_dtd=True,
        forbid_entities=True,
        forbid_external=True,
    )

    # Not allowing DTD, do a primitive strip, and allow parsing to fail if it was mangled.
    if isinstance(xml_string, str) and xml_string.startswith("<?"):
        xml_string = xml_string[xml_string.find("?>") + 2 :]

    try:
        parser.feed(xml_string)
        return parser.close()
    except ParseError as e:
        # Offer consistent results for callers to check for invalid data.
        logger.debug("Parsing XML error: %s: %s", e, xml_string)
        raise ExternalParsingError(str(e)) from e


def split_ns(xml_name: str) -> tuple[str | None, str]:
    """Split the element tag or attribute/text value into the namespace and
    local name. The stdlib etree doesn't have the properties for this (lxml does).
    """
    # Tags may start with a `{ns}`
    if xml_name.startswith("{"):
        end = xml_name.index("}")
        return xml_name[1:end], xml_name[end + 1 :]
    else:
        return None, xml_name
